#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Domain classification"""

from typing import Optional
from typing import Text

import numpy as np
from .base import LabelingTask
from .base import LabelingTaskGenerator
from pyannote.audio.train.task import Task, TaskType, TaskOutput
from pyannote.audio.features.wrapper import Wrappable
from pyannote.database import Protocol
from pyannote.database import Subset
from pyannote.audio.train.model import Resolution
from pyannote.audio.train.model import Alignment


class DomainClassificationGenerator(LabelingTaskGenerator):
    """Batch generator for training domain classification

    Parameters
    ----------
    task : Task
        Task
    feature_extraction : Wrappable
        Describes how features should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
    protocol : Protocol
    subset : {'train', 'development', 'test'}, optional
        Protocol and subset.
    resolution : `pyannote.core.SlidingWindow`, optional
        Override `feature_extraction.sliding_window`. This is useful for
        models that include the feature extraction step (e.g. SincNet) and
        therefore output a lower sample rate than that of the input.
        Defaults to `feature_extraction.sliding_window`
    alignment : {'center', 'loose', 'strict'}, optional
        Which mode to use when cropping labels. This is useful for models that
        include the feature extraction step (e.g. SincNet) and therefore use a
        different cropping mode. Defaults to 'center'.
    duration : float, optional
        Duration of audio chunks. Defaults to 2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Force total audio duration per epoch, in days.
        Defaults to total duration of protocol subset.
    domain : `str`, optional
        Key to use as domain. Defaults to 'domain'.
    """

    def __init__(
        self,
        task: Task,
        feature_extraction: Wrappable,
        protocol: Protocol,
        subset: Subset = "train",
        resolution: Optional[Resolution] = None,
        alignment: Optional[Alignment] = None,
        duration: float = 2.0,
        batch_size: int = 32,
        per_epoch: float = None,
        domain: Text = "domain",
    ):

        self.domain = domain

        super().__init__(
            task,
            feature_extraction,
            protocol,
            subset=subset,
            resolution=resolution,
            alignment=alignment,
            duration=duration,
            batch_size=batch_size,
            per_epoch=per_epoch,
            exhaustive=False,
        )

    def initialize_y(self, current_file):
        return self.file_labels_[self.domain].index(current_file[self.domain])

    def crop_y(self, y, segment):
        return y

    @property
    def specifications(self):
        return {
            "task": self.task,
            "X": {"dimension": self.feature_extraction.dimension},
            "y": {"classes": self.file_labels_[self.domain]},
        }


class DomainClassification(LabelingTask):
    """Train domain classification

    Parameters
    ----------
    domain : `str`, optional
        Key to use as domain. Defaults to 'domain'.
    duration : float, optional
        Duration of sub-sequences. Defaults to 3.2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Total audio duration per epoch, in days.
        Defaults to one day (1).
    """

    def __init__(self, domain="domain", **kwargs):
        super().__init__(**kwargs)
        self.domain = domain

    def get_batch_generator(
        self,
        feature_extraction: Wrappable,
        protocol: Protocol,
        subset: Subset = "train",
        **kwargs
    ) -> DomainClassificationGenerator:
        """Get batch generator for domain classification

        Parameters
        ----------
        feature_extraction : Wrappable
            Feature extraction.
        protocol : Protocol
        subset : {'train', 'development', 'test'}, optional
            Protocol and subset used for batch generation.

        Returns
        -------
        batch_generator : `DomainClassificationGenerator`
            Batch generator
        """
        return DomainClassificationGenerator(
            self.task,
            feature_extraction,
            protocol,
            subset=subset,
            domain=self.domain,
            duration=self.duration,
            per_epoch=self.per_epoch,
            batch_size=self.batch_size,
        )

    @property
    def task(self):
        return Task(type=TaskType.MULTI_CLASS_CLASSIFICATION, output=TaskOutput.VECTOR)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import torch
import torch.nn as nn
from typing import List


class Linear(nn.Module):
    """Linear layers

    Parameters
    ----------
    n_features : int
        Input feature shape.
    hidden_size : list of int, optional
        Number of features in hidden. Defaults to [256, 128].
    bias : bool, optional
        If set to False, the layer will not learn an additive bias.
    """

    def __init__(
        self, n_features: int, hidden_size: List[int] = [256, 128], bias: bool = True,
    ):
        super().__init__()

        self.n_features = n_features
        self.hidden_size = hidden_size
        self.bias = bias

        self.linear = nn.ModuleList()

        for out_features in hidden_size:
            linear = nn.Linear(n_features, out_features, bias=self.bias)
            self.linear.append(linear)
            n_features = out_features

        self.activation = nn.Tanh()

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """Forward pass

        Parameter
        ---------
        features : torch.Tensor
            Feature tensor with shape (batch_size, n_frames, n_features) or
            (batch_size, n_features).

        Returns
        -------
        output : torch.Tensor
            Output features with shape (batch_size, n_frames, hidden_size[-1])
            or (batch_size, hidden_size[-1])
        """

        output = features
        for linear in self.linear:
            output = linear(output)
            output = self.activation(output)
        return output

    @property
    def num_layers(self) -> int:
        """Number of linear layers"""
        return len(self.hidden_size)

    @property
    def dimension(self) -> int:
        """Dimension of output features"""
        return self.hidden_size[-1]



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Speaker change detection"""

from typing import Optional
from typing import Text

import numpy as np
from .base import LabelingTask
from .base import LabelingTaskGenerator
from pyannote.audio.train.task import Task, TaskType, TaskOutput
import scipy.signal
from pyannote.audio.features.wrapper import Wrappable
from pyannote.database import Protocol
from pyannote.database import Subset
from pyannote.audio.train.model import Resolution
from pyannote.audio.train.model import Alignment
from pyannote.audio.train.model import RESOLUTION_FRAME


class SpeakerChangeDetectionGenerator(LabelingTaskGenerator):
    """Batch generator for training speaker change detection

    Parameters
    ----------
    feature_extraction : Wrappable
        Describes how features should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
    protocol : Protocol
    subset : {'train', 'development', 'test'}, optional
        Protocol and subset.
    resolution : `pyannote.core.SlidingWindow`, optional
        Override `feature_extraction.sliding_window`. This is useful for
        models that include the feature extraction step (e.g. SincNet) and
        therefore output a lower sample rate than that of the input.
        Defaults to `feature_extraction.sliding_window`
    alignment : {'center', 'loose', 'strict'}, optional
        Which mode to use when cropping labels. This is useful for models that
        include the feature extraction step (e.g. SincNet) and therefore use a
        different cropping mode. Defaults to 'center'.
    duration : float, optional
        Duration of audio chunks. Defaults to 2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Force total audio duration per epoch, in days.
        Defaults to total duration of protocol subset.
    mask : str, optional
        When provided, protocol files are expected to contain a key named after
        this `mask` variable and providing a `SlidingWindowFeature` instance.
        Generated batches will contain an additional "mask" key (on top of
        existing "X" and "y" keys) computed as an excerpt of `current_file[mask]`
        time-aligned with "y". Defaults to not add any "mask" key.
    collar : float, optional
        Duration of positive collar, in seconds. Default to 0.1 (i.e. frames
        less than 100ms away from the actual change are also labeled as
        change).
    regression : bool, optional
        Use triangle-shaped label sequences centered on actual changes.
        Defaults to False (i.e. rectangle-shaped label sequences).
    non_speech : bool, optional
        Keep non-speech/speaker changes. Defauls to False (i.e. only
        speaker/speaker changes are marked as such).
    """

    def __init__(
        self,
        task: Task,
        feature_extraction: Wrappable,
        protocol: Protocol,
        subset: Subset = "train",
        resolution: Optional[Resolution] = None,
        alignment: Optional[Alignment] = None,
        duration: float = 2.0,
        batch_size: int = 32,
        per_epoch: float = None,
        mask: Text = None,
        collar: float = 0.1,
        regression: bool = False,
        non_speech: bool = False,
    ):

        self.collar = collar
        self.regression = regression
        self.non_speech = non_speech

        # number of samples in collar
        if resolution in [None, RESOLUTION_FRAME]:
            resolution = feature_extraction.sliding_window
        self.collar_ = resolution.duration_to_samples(collar)

        # window
        self.window_ = scipy.signal.triang(self.collar_)[:, np.newaxis]

        super().__init__(
            task,
            feature_extraction,
            protocol,
            subset=subset,
            resolution=resolution,
            alignment=alignment,
            duration=duration,
            batch_size=batch_size,
            per_epoch=per_epoch,
            exhaustive=False,
            mask=mask,
            local_labels=True,
        )

    def postprocess_y(self, Y: np.ndarray) -> np.ndarray:
        """Generate labels for speaker change detection

        Parameters
        ----------
        Y : (n_samples, n_speakers) numpy.ndarray
            Discretized annotation returned by `pyannote.core.utils.numpy.one_hot_encoding`.

        Returns
        -------
        y : (n_samples, 1) numpy.ndarray

        See also
        --------
        `pyannote.core.utils.numpy.one_hot_encoding`
        """

        # replace NaNs by 0s
        Y = np.nan_to_num(Y)
        n_samples, n_speakers = Y.shape

        # True = change. False = no change
        y = np.sum(np.abs(np.diff(Y, axis=0)), axis=1, keepdims=True)
        y = np.vstack(([[0]], y > 0))

        # mark change points neighborhood as positive
        y = np.minimum(1, scipy.signal.convolve(y, self.window_, mode="same"))

        # HACK for some reason, y rarely equals zero
        if not self.regression:
            y = 1 * (y > 1e-10)

        # at this point, all segment boundaries are marked as change
        # (including non-speech/speaker changesà

        # remove non-speech/speaker change
        if not self.non_speech:

            # append (half collar) empty samples at the beginning/end
            expanded_Y = np.vstack(
                [
                    np.zeros(((self.collar_ + 1) // 2, n_speakers), dtype=Y.dtype),
                    Y,
                    np.zeros(((self.collar_ + 1) // 2, n_speakers), dtype=Y.dtype),
                ]
            )

            # stride trick. data[i] is now a sliding window of collar length
            # centered at time step i.
            data = np.lib.stride_tricks.as_strided(
                expanded_Y,
                shape=(n_samples, n_speakers, self.collar_),
                strides=(Y.strides[0], Y.strides[1], Y.strides[0]),
            )

            # y[i] = 1 if more than one speaker are speaking in the
            # corresponding window. 0 otherwise
            x_speakers = 1 * (np.sum(np.sum(data, axis=2) > 0, axis=1) > 1)
            x_speakers = x_speakers.reshape(-1, 1)

            y *= x_speakers

        return y

    @property
    def specifications(self):

        specs = {
            "task": self.task,
            "X": {"dimension": self.feature_extraction.dimension},
        }

        if self.regression:
            specs["y"] = {"classes": ["change",]}
        else:
            specs["y"] = {"classes": ["non_change", "change"]}

        return specs


class SpeakerChangeDetection(LabelingTask):
    """Train speaker change detection

    Parameters
    ----------
    collar : float, optional
        Duration of positive collar, in seconds. Default to 0.1 (i.e. frames
        less than 100ms away from the actual change are also labeled as
        change).
    regression : bool, optional
        Use triangle-shaped label sequences centered on actual changes.
        Defaults to False (i.e. rectangle-shaped label sequences).
    non_speech : bool, optional
        Keep non-speech/speaker changes (and vice-versa). Defauls to False
        (i.e. only keep speaker/speaker changes).
    duration : float, optional
        Duration of sub-sequences. Defaults to 3.2s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    per_epoch : float, optional
        Total audio duration per epoch, in days.
        Defaults to one day (1).
    """

    def __init__(self, collar=0.100, regression=False, non_speech=False, **kwargs):
        super().__init__(**kwargs)
        self.collar = collar
        self.regression = regression
        self.non_speech = non_speech

    def get_batch_generator(
        self,
        feature_extraction,
        protocol,
        subset: Subset = "train",
        resolution=None,
        alignment=None,
    ):
        """
        resolution : `pyannote.core.SlidingWindow`, optional
            Override `feature_extraction.sliding_window`. This is useful for
            models that include the feature extraction step (e.g. SincNet) and
            therefore output a lower sample rate than that of the input.
        alignment : {'center', 'loose', 'strict'}, optional
            Which mode to use when cropping labels. This is useful for models
            that include the feature extraction step (e.g. SincNet) and
            therefore use a different cropping mode. Defaults to 'center'.
        """
        return SpeakerChangeDetectionGenerator(
            self.task,
            feature_extraction,
            protocol,
            resolution=resolution,
            alignment=alignment,
            subset=subset,
            collar=self.collar,
            regression=self.regression,
            non_speech=self.non_speech,
            duration=self.duration,
            batch_size=self.batch_size,
            per_epoch=self.per_epoch,
        )

    @property
    def task(self):

        if self.regression:
            return Task(type=TaskType.REGRESSION, output=TaskOutput.SEQUENCE)
        else:
            return Task(
                type=TaskType.MULTI_CLASS_CLASSIFICATION, output=TaskOutput.SEQUENCE
            )



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
TODO
"""

from abc import ABCMeta, abstractmethod
from typing import Iterator

from pyannote.audio.features.base import FeatureExtraction
from pyannote.database import Protocol
from pyannote.database import Subset

import warnings
import numpy as np
import pescador


class BatchGenerator(metaclass=ABCMeta):
    """Batch generator base class

    Parameters
    ----------
    feature_extraction : `FeatureExtraction`
    protocol : `Protocol`
        pyannote.database protocol used by the generator.
    subset : {'train', 'development', 'test'}, optional
        Subset used by the generator. Defaults to 'train'.
    """

    @abstractmethod
    def __init__(
        self,
        feature_extraction: FeatureExtraction,
        protocol: Protocol,
        subset: Subset = "train",
        **kwargs,
    ):
        pass

    @property
    @abstractmethod
    def specifications(self) -> dict:
        """Generator specifications

        Returns
        -------
        specifications : `dict`
            Dictionary describing generator specifications.
        """
        pass

    @property
    @abstractmethod
    def batches_per_epoch(self) -> int:
        """Number of batches per epoch

        Returns
        -------
        n_batches : `int`
            Number of batches to make an epoch.
        """
        pass

    @abstractmethod
    def samples(self) -> Iterator:
        pass

    def __call__(self) -> Iterator:
        batches = pescador.maps.buffer_stream(
            self.samples(), self.batch_size, partial=False, axis=None
        )

        while True:
            next_batch = next(batches)
            # HACK in some rare cases, .samples() yields samples
            # HACK with different length leading to batch being of
            # HACK type "object". for now, we simply discard those
            # HACK buggy batches.
            # TODO fix the problem upstream in .samples()
            if any(batch.dtype == np.object_ for batch in next_batch.values()):
                msg = f"Skipping malformed batch."
                warnings.warn(msg)
                continue
            yield next_batch



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from typing import Text, Union, Tuple, List
from pathlib import Path
from pyannote.core import (
    Segment,
    SlidingWindow,
    Timeline,
    Annotation,
    SlidingWindowFeature,
)
from pyannote.database import ProtocolFile

import numpy as np

from pyannote.pipeline import Pipeline
from pyannote.pipeline.parameter import Uniform

from pyannote.audio.features.wrapper import Wrapper
from pyannote.audio.features.utils import get_audio_duration

from pyannote.metrics.diarization import DiarizationErrorRate
from pyannote.metrics.detection import DetectionErrorRate

from pyannote.audio.utils.signal import Binarize

from pyannote.core.utils.hierarchy import pool
from scipy.cluster.hierarchy import fcluster
from scipy.spatial.distance import cdist


from .utils import time2index
from .utils import index2index

# Hyper-parameters tuned to minimize diarization error rate
# on one third of DIHARD II training set
PRETRAINED_PARAMS = {
    "emb_duration": 1.7657045140297274,
    "emb_step_ratio": 0.20414598809353782,
    "emb_threshold": 0.5274911675340328,
    "sad_min_duration_off": 0.13583405625051126,
    "sad_min_duration_on": 0.0014190874731107286,
    "sad_threshold_off": 0.7878607185085043,
    "sad_threshold_on": 0.5940560764213958,
}


class InteractiveDiarization(Pipeline):
    """Interactive diarization pipeline

    Parameters
    ----------
    sad : str or Path, optional
        Pretrained speech activity detection model. Defaults to "sad".
    emb : str or Path, optional
        Pretrained speaker embedding model. Defaults to "emb".
    batch_size : int, optional
        Batch size.
    only_sad : bool, optional
        Set to True if you only care about speech activity detection.

    Hyper-parameters
    ----------------
    sad_threshold_on, sad_threshold_off : float
        Onset/offset speech activity detection thresholds.
    sad_min_duration_on, sad_min_duration_off : float
        Minimum duration of speech/non-speech regions.
    emb_duration, emb_step_ratio : float
        Sliding window used for embedding extraction.
    emb_threshold : float
        Distance threshold used as stopping criterion for hierarchical
        agglomerative clustering.
    """

    def __init__(
        self,
        sad: Union[Text, Path] = {"sad": {"duration": 2.0, "step": 0.1}},
        emb: Union[Text, Path] = "emb",
        batch_size: int = None,
        only_sad: bool = False,
    ):

        super().__init__()

        self.sad = Wrapper(sad)
        if batch_size is not None:
            self.sad.batch_size = batch_size
        self.sad_speech_index_ = self.sad.classes.index("speech")

        self.sad_threshold_on = Uniform(0.0, 1.0)
        self.sad_threshold_off = Uniform(0.0, 1.0)
        self.sad_min_duration_on = Uniform(0.0, 0.5)
        self.sad_min_duration_off = Uniform(0.0, 0.5)

        self.only_sad = only_sad
        if self.only_sad:
            return

        self.emb = Wrapper(emb)
        if batch_size is not None:
            self.emb.batch_size = batch_size

        max_duration = self.emb.duration
        min_duration = getattr(self.emb, "min_duration", 0.25 * max_duration)
        self.emb_duration = Uniform(min_duration, max_duration)
        self.emb_step_ratio = Uniform(0.1, 1.0)
        self.emb_threshold = Uniform(0.0, 2.0)

    def initialize(self):
        """Initialize pipeline internals with current hyper-parameter values"""

        self.sad_binarize_ = Binarize(
            onset=self.sad_threshold_on,
            offset=self.sad_threshold_off,
            min_duration_on=self.sad_min_duration_on,
            min_duration_off=self.sad_min_duration_off,
        )

        if not self.only_sad:
            # embeddings will be extracted with a sliding window
            # of "emb_duration" duration and "emb_step_ratio x emb_duration" step.
            self.emb.duration = self.emb_duration
            self.emb.step = self.emb_step_ratio

    def compute_speech(self, current_file: ProtocolFile) -> Timeline:
        """Apply speech activity detection

        Parameters
        ----------
        current_file : ProtocolFile
            Protocol file.

        Returns
        -------
        speech : Timeline
            Speech activity detection result.
        """

        # speech activity detection
        if "sad_scores" in current_file:
            sad_scores: SlidingWindowFeature = current_file["sad_scores"]
        else:
            sad_scores = self.sad(current_file)
            if np.nanmean(sad_scores) < 0:
                sad_scores = np.exp(sad_scores)
            current_file["sad_scores"] = sad_scores

        speech: Timeline = self.sad_binarize_.apply(
            sad_scores, dimension=self.sad_speech_index_
        )

        return speech

    def compute_embedding(self, current_file: ProtocolFile) -> SlidingWindowFeature:
        """Extract speaker embedding

        Parameters
        ----------
        current_file : ProtocolFile
            Protocol file

        Returns
        -------
        embedding : SlidingWindowFeature
            Speaker embedding.
        """

        return self.emb(current_file)

    def get_segment_assignment(
        self, embedding: SlidingWindowFeature, speech: Timeline
    ) -> np.ndarray:
        """Get segment assignment

        Parameters
        ----------
        embedding : SlidingWindowFeature
            Embeddings.
        speech : Timeline
            Speech regions.

        Returns
        -------
        assignment : (num_embedding, ) np.ndarray
            * assignment[i] = s with s > 0 means that ith embedding is strictly
            contained in (1-based) sth segment.
            * assignment[i] = s with s < 0 means that more than half of ith
            embedding is part of (1-based) sth segment.
            * assignment[i] = 0 means that none of the above is true.
        """

        assignment: np.ndarray = np.zeros((len(embedding),), dtype=np.int32)

        for s, segment in enumerate(speech):
            indices = embedding.sliding_window.crop(segment, mode="strict")
            if len(indices) > 0:
                strict = 1
            else:
                strict = -1
                indices = embedding.sliding_window.crop(segment, mode="center")
            for i in indices:
                if i < 0 or i >= len(embedding):
                    continue
                assignment[i] = strict * (s + 1)

        return assignment

    def __call__(
        self,
        current_file: ProtocolFile,
        cannot_link: List[Tuple[float, float]] = None,
        must_link: List[Tuple[float, float]] = None,
    ) -> Annotation:
        """Apply speaker diarization

        Parameters
        ----------
        current_file : ProtocolFile
            Protocol file.
        cannot_link :
            List of time-based "cannot link" constraints.
        must_link :
            List of time-based "must link" constraints.

        Returns
        -------
        diarization : Annotation
            Speaker diarization result.
        """

        if cannot_link is None:
            cannot_link = []
        if must_link is None:
            must_link = []

        if "duration" not in current_file:
            current_file["duration"] = get_audio_duration(current_file)

        # in "interactive annotation" mode, there is no need to recompute speech
        # regions every time a file is processed: they can be passed with the
        # file directly
        if "speech" in current_file:
            speech: Timeline = current_file["speech"]

        # in "pipeline optimization" mode, pipeline hyper-parameters are different
        # every time a file is processed: speech regions must be recomputed
        else:
            speech = self.compute_speech(current_file)

        if self.only_sad:
            return speech.to_annotation(generator=iter(lambda: "SPEECH", None))

        # in "interactive annotation" mode, pipeline hyper-parameters are fixed.
        # therefore, there is no need to recompute embeddings every time a file
        # is processed: they can be passed with the file directly.
        if "embedding" in current_file:
            embedding: SlidingWindowFeature = current_file["embedding"]

        # in "pipeline optimization" mode, pipeline hyper-parameters are different
        # every time a file is processed: embeddings must be recomputed
        else:
            embedding = self.compute_embedding(current_file)

        window: SlidingWindow = embedding.sliding_window

        # segment_assignment[i] = s with s > 0 means that ith embedding is
        #       strictly contained in (1-based) sth segment.
        # segment_assignment[i] = s with s < 0 means that more than half of ith
        #       embedding is part of (1-based) sth segment.
        # segment_assignment[i] = 0 means that none of the above is true.
        segment_assignment: np.ndarray = self.get_segment_assignment(embedding, speech)

        # cluster_assignment[i] = k (k > 0) means that the ith embedding belongs
        #                           to kth cluster
        # cluster_assignment[i] = 0 when segment_assignment[i] = 0
        cluster_assignment: np.ndarray = np.zeros((len(embedding),), dtype=np.int32)

        clean = segment_assignment > 0
        noisy = segment_assignment < 0
        clean_indices = np.where(clean)[0]
        if len(clean_indices) < 2:
            cluster_assignment[clean_indices] = 1

        else:

            # convert time-based constraints to index-based constraints
            cannot_link = index2index(time2index(cannot_link, window), clean)
            must_link = index2index(time2index(must_link, window), clean)

            dendrogram = pool(
                embedding[clean_indices],
                metric="cosine",
                cannot_link=cannot_link,
                must_link=must_link,
                must_link_method="propagate",
            )
            clusters = fcluster(dendrogram, self.emb_threshold, criterion="distance")
            for i, k in zip(clean_indices, clusters):
                cluster_assignment[i] = k

        loose_indices = np.where(noisy)[0]
        if len(clean_indices) == 0:
            if len(loose_indices) < 2:
                clusters = [1] * len(loose_indices)
            else:
                dendrogram = pool(embedding[loose_indices], metric="cosine")
                clusters = fcluster(
                    dendrogram, self.emb_threshold, criterion="distance"
                )
            for i, k in zip(loose_indices, clusters):
                cluster_assignment[i] = k

        else:
            # NEAREST NEIGHBOR
            distance = cdist(
                embedding[clean_indices], embedding[loose_indices], metric="cosine"
            )
            nearest_neighbor = np.argmin(distance, axis=0)
            for loose_index, nn in zip(loose_indices, nearest_neighbor):
                strict_index = clean_indices[nn]
                cluster_assignment[loose_index] = cluster_assignment[strict_index]

            # # NEAREST CLUSTER
            # centroid = np.vstack(
            #     [
            #         np.mean(embedding[cluster_assignment == k], axis=0)
            #         for k in np.unique(clusters)
            #     ]
            # )
            # distance = cdist(centroid, embedding[loose_indices], metric="cosine")
            # cluster_assignment[loose_indices] = np.argmin(distance, axis=0) + 1

        # convert cluster assignment to pyannote.core.Annotation
        # (make sure to keep speech regions unchanged)
        hypothesis = Annotation(uri=current_file.get("uri", None))
        for s, segment in enumerate(speech):

            indices = np.where(segment_assignment == s + 1)[0]
            if len(indices) == 0:
                indices = np.where(segment_assignment == -(s + 1))[0]
                if len(indices) == 0:
                    continue

            clusters = cluster_assignment[indices]

            start, k = segment.start, clusters[0]
            change_point = np.diff(clusters) != 0
            for i, new_k in zip(indices[1:][change_point], clusters[1:][change_point]):
                end = window[i].middle + 0.5 * window.step
                hypothesis[Segment(start, end)] = k
                start = end
                k = new_k
            hypothesis[Segment(start, segment.end)] = k

        return hypothesis.support()

    def get_metric(self) -> Union[DetectionErrorRate, DiarizationErrorRate]:
        if self.only_sad:
            return DetectionErrorRate(collar=0.0)
        else:
            return DiarizationErrorRate(collar=0.0, skip_overlap=False)



# The MIT License (MIT)
#
# Copyright (c) 2019 Mirco Ravanelli
# Copyright (c) 2019-2020 CNRS
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# AUTHOR
# Hervé Bredin - http://herve.niderb.fr

# Part of this code was taken from https://github.com/mravanelli/SincNet
# (see above license terms).

# Please give proper credit to the authors if you are using SincNet-based
# models  by citing their paper:

# Mirco Ravanelli, Yoshua Bengio.
# "Speaker Recognition from raw waveform with SincNet".
# SLT 2018. https://arxiv.org/abs/1808.00158

from typing import List
import numpy as np
import torch
import torch.nn.functional as F
import torch.nn as nn
import math
from pyannote.core import SlidingWindow
from pyannote.audio.train.task import Task


class SincConv1d(nn.Module):
    """Sinc-based 1D convolution

    Parameters
    ----------
    in_channels : `int`
        Should be 1.
    out_channels : `int`
        Number of filters.
    kernel_size : `int`
        Filter length.
    stride : `int`, optional
        Defaults to 1.
    sample_rate : `int`, optional
        Sample rate. Defaults to 16000.
    min_low_hz: `int`, optional
        Defaults to 50.
    min_band_hz: `int`, optional
        Defaults to 50.

    Usage
    -----
    Same as `torch.nn.Conv1d`

    Reference
    ---------
    Mirco Ravanelli, Yoshua Bengio. "Speaker Recognition from raw waveform with
    SincNet". SLT 2018. https://arxiv.org/abs/1808.00158
    """

    @staticmethod
    def to_mel(hz):
        return 2595 * np.log10(1 + hz / 700)

    @staticmethod
    def to_hz(mel):
        return 700 * (10 ** (mel / 2595) - 1)

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        sample_rate=16000,
        min_low_hz=50,
        min_band_hz=50,
        stride=1,
        padding=0,
        dilation=1,
        bias=False,
        groups=1,
    ):

        super().__init__()

        if in_channels != 1:
            msg = (
                f"SincConv1d only supports one input channel. "
                f"Here, in_channels = {in_channels}."
            )
            raise ValueError(msg)
        self.in_channels = in_channels

        self.out_channels = out_channels

        if kernel_size % 2 == 0:
            msg = (
                f"SincConv1d only support odd kernel size. "
                f"Here, kernel_size = {kernel_size}."
            )
            raise ValueError(msg)
        self.kernel_size = kernel_size

        self.stride = stride
        self.padding = padding
        self.dilation = dilation

        if bias:
            raise ValueError("SincConv1d does not support bias.")
        if groups > 1:
            raise ValueError("SincConv1d does not support groups.")

        self.sample_rate = sample_rate
        self.min_low_hz = min_low_hz
        self.min_band_hz = min_band_hz

        # initialize filterbanks such that they are equally spaced in Mel scale
        low_hz = 30
        high_hz = self.sample_rate / 2 - (self.min_low_hz + self.min_band_hz)

        mel = np.linspace(
            self.to_mel(low_hz), self.to_mel(high_hz), self.out_channels + 1
        )
        hz = self.to_hz(mel)

        # filter lower frequency (out_channels, 1)
        self.low_hz_ = nn.Parameter(torch.Tensor(hz[:-1]).view(-1, 1))

        # filter frequency band (out_channels, 1)
        self.band_hz_ = nn.Parameter(torch.Tensor(np.diff(hz)).view(-1, 1))

        # Half Hamming half window
        n_lin = torch.linspace(
            0, self.kernel_size / 2 - 1, steps=int((self.kernel_size / 2))
        )
        self.window_ = 0.54 - 0.46 * torch.cos(2 * math.pi * n_lin / self.kernel_size)

        # (kernel_size, 1)
        # Due to symmetry, I only need half of the time axes
        n = (self.kernel_size - 1) / 2.0
        self.n_ = 2 * math.pi * torch.arange(-n, 0).view(1, -1) / self.sample_rate

    def forward(self, waveforms):
        """Get sinc filters activations

        Parameters
        ----------
        waveforms : `torch.Tensor` (batch_size, 1, n_samples)
            Batch of waveforms.

        Returns
        -------
        features : `torch.Tensor` (batch_size, out_channels, n_samples_out)
            Batch of sinc filters activations.
        """

        self.n_ = self.n_.to(waveforms.device)
        self.window_ = self.window_.to(waveforms.device)

        low = self.min_low_hz + torch.abs(self.low_hz_)

        high = torch.clamp(
            low + self.min_band_hz + torch.abs(self.band_hz_),
            self.min_low_hz,
            self.sample_rate / 2,
        )
        band = (high - low)[:, 0]

        f_times_t_low = torch.matmul(low, self.n_)
        f_times_t_high = torch.matmul(high, self.n_)

        # Equivalent to Eq.4 of the reference paper
        # I just have expanded the sinc and simplified the terms.
        # This way I avoid several useless computations.
        band_pass_left = (
            (torch.sin(f_times_t_high) - torch.sin(f_times_t_low)) / (self.n_ / 2)
        ) * self.window_
        band_pass_center = 2 * band.view(-1, 1)
        band_pass_right = torch.flip(band_pass_left, dims=[1])

        band_pass = torch.cat(
            [band_pass_left, band_pass_center, band_pass_right], dim=1
        )

        band_pass = band_pass / (2 * band[:, None])

        self.filters = (band_pass).view(self.out_channels, 1, self.kernel_size)

        return F.conv1d(
            waveforms,
            self.filters,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            bias=None,
            groups=1,
        )


class SincNet(nn.Module):
    """SincNet (learnable) feature extraction

    Parameters
    ----------
    waveform_normalize : `bool`, optional
        Standardize waveforms (to zero mean and unit standard deviation) and
        apply (learnable) affine transform. Defaults to True.
    instance_normalize : `bool`, optional
        Standardize internal representation (to zero mean and unit standard
        deviation) and apply (learnable) affine transform. Defaults to True.


    Reference
    ---------
    Mirco Ravanelli, Yoshua Bengio. "Speaker Recognition from raw waveform with
    SincNet". SLT 2018. https://arxiv.org/abs/1808.00158

    """

    @staticmethod
    def get_alignment(task: Task, **kwargs):
        """Get frame alignment"""
        return "strict"

    @staticmethod
    def get_resolution(
        task: Task,
        sample_rate: int = 16000,
        kernel_size: List[int] = [251, 5, 5],
        stride: List[int] = [1, 1, 1],
        max_pool: List[int] = [3, 3, 3],
        **kwargs,
    ) -> SlidingWindow:
        """Get frame resolution

        Parameters
        ----------
        task : Task
        sample_rate : int, optional
        kerne_size : list of int, optional
        stride : list of int, optional
        max_pool : list of int, optional

        Returns
        -------
        resolution : SlidingWindow
            Frame resolution.
        """

        # https://medium.com/mlreview/a-guide-to-receptive-field-arithmetic-for-convolutional-neural-networks-e0f514068807
        padding = 0
        receptive_field, jump, start = 1, 1, 0.5
        for ks, s, mp in zip(kernel_size, stride, max_pool):
            # increase due to (Sinc)Conv1d
            receptive_field += (ks - 1) * jump
            start += ((ks - 1) / 2 - padding) * jump
            jump *= s
            # increase in receptive field due to MaxPool1d
            receptive_field += (mp - 1) * jump
            start += ((mp - 1) / 2 - padding) * jump
            jump *= mp

        return SlidingWindow(
            duration=receptive_field / sample_rate, step=jump / sample_rate, start=0.0
        )

    def __init__(
        self,
        waveform_normalize=True,
        sample_rate=16000,
        min_low_hz=50,
        min_band_hz=50,
        out_channels=[80, 60, 60],
        kernel_size: List[int] = [251, 5, 5],
        stride=[1, 1, 1],
        max_pool=[3, 3, 3],
        instance_normalize=True,
        activation="leaky_relu",
        dropout=0.0,
    ):
        super().__init__()

        # check parameters values
        n_layers = len(out_channels)
        if len(kernel_size) != n_layers:
            msg = (
                f"out_channels ({len(out_channels):d}) and kernel_size "
                f"({len(kernel_size):d}) should have the same length."
            )
            raise ValueError(msg)
        if len(stride) != n_layers:
            msg = (
                f"out_channels ({len(out_channels):d}) and stride "
                f"({len(stride):d}) should have the same length."
            )
            raise ValueError(msg)
        if len(max_pool) != n_layers:
            msg = (
                f"out_channels ({len(out_channels):d}) and max_pool "
                f"({len(max_pool):d}) should have the same length."
            )
            raise ValueError(msg)

        # Waveform normalization
        self.waveform_normalize = waveform_normalize
        if self.waveform_normalize:
            self.waveform_normalize_ = torch.nn.InstanceNorm1d(1, affine=True)

        # SincNet-specific parameters
        self.sample_rate = sample_rate
        self.min_low_hz = min_low_hz
        self.min_band_hz = min_band_hz

        # Conv1D parameters
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.conv1d_ = nn.ModuleList([])

        # Max-pooling parameters
        self.max_pool = max_pool
        self.max_pool1d_ = nn.ModuleList([])

        # Instance normalization
        self.instance_normalize = instance_normalize
        if self.instance_normalize:
            self.instance_norm1d_ = nn.ModuleList([])

        config = zip(self.out_channels, self.kernel_size, self.stride, self.max_pool)

        in_channels = None
        for i, (out_channels, kernel_size, stride, max_pool) in enumerate(config):

            # 1D convolution
            if i > 0:
                conv1d = nn.Conv1d(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=stride,
                    padding=0,
                    dilation=1,
                    groups=1,
                    bias=True,
                )
            else:
                conv1d = SincConv1d(
                    1,
                    out_channels,
                    kernel_size,
                    sample_rate=self.sample_rate,
                    min_low_hz=self.min_low_hz,
                    min_band_hz=self.min_band_hz,
                    stride=stride,
                    padding=0,
                    dilation=1,
                    bias=False,
                    groups=1,
                )
            self.conv1d_.append(conv1d)

            # 1D max-pooling
            max_pool1d = nn.MaxPool1d(max_pool, stride=max_pool, padding=0, dilation=1)
            self.max_pool1d_.append(max_pool1d)

            # 1D instance normalization
            if self.instance_normalize:
                instance_norm1d = nn.InstanceNorm1d(out_channels, affine=True)
                self.instance_norm1d_.append(instance_norm1d)

            in_channels = out_channels

        # Activation function
        self.activation = activation
        if self.activation == "leaky_relu":
            self.activation_ = nn.LeakyReLU(negative_slope=0.2)
        else:
            msg = f'Only "leaky_relu" activation is supported.'
            raise ValueError(msg)

        # Dropout
        self.dropout = dropout
        if self.dropout:
            self.dropout_ = nn.Dropout(p=self.dropout)

    def forward(self, waveforms):
        """Extract SincNet features

        Parameters
        ----------
        waveforms : (batch_size, n_samples, 1)
            Batch of waveforms

        Returns
        -------
        features : (batch_size, n_frames, out_channels[-1])
        """

        output = waveforms.transpose(1, 2)

        # standardize waveforms
        if self.waveform_normalize:
            output = self.waveform_normalize_(output)

        layers = zip(self.conv1d_, self.max_pool1d_)
        for i, (conv1d, max_pool1d) in enumerate(layers):

            output = conv1d(output)
            if i == 0:
                output = torch.abs(output)

            output = max_pool1d(output)

            if self.instance_normalize:
                output = self.instance_norm1d_[i](output)

            output = self.activation_(output)

            if self.dropout:
                output = self.dropout_(output)

        return output.transpose(1, 2)

    def dimension():
        doc = "Output features dimension."

        def fget(self):
            return self.out_channels[-1]

        return locals()

    dimension = property(**dimension())



#!/usr/bin/env python
# encoding: utf-8

# Shamelessly stolen from
# https://gist.github.com/TySkby/143190ad1b88c6115597c45f996b030c

"""Easily put time restrictions on things

Note: Requires Python 3.x

Usage as a context manager:
```
with timeout(10):
    something_that_should_not_exceed_ten_seconds()
```

Usage as a decorator:
```
@timeout(10)
def something_that_should_not_exceed_ten_seconds():
    do_stuff_with_a_timeout()
```

Handle timeouts:
```
try:
   with timeout(10):
       something_that_should_not_exceed_ten_seconds()
   except TimeoutError:
       log('Got a timeout, couldn't finish')
```

Suppress TimeoutError and just die after expiration:
```
with timeout(10, suppress_timeout_errors=True):
    something_that_should_not_exceed_ten_seconds()

print('Maybe exceeded 10 seconds, but finished either way')
```
"""
import contextlib
import errno
import os
import signal


DEFAULT_TIMEOUT_MESSAGE = os.strerror(errno.ETIME)


class timeout(contextlib.ContextDecorator):
    def __init__(
        self,
        seconds,
        *,
        timeout_message=DEFAULT_TIMEOUT_MESSAGE,
        suppress_timeout_errors=False
    ):
        self.seconds = int(seconds)
        self.timeout_message = timeout_message
        self.suppress = bool(suppress_timeout_errors)

    def _timeout_handler(self, signum, frame):
        raise TimeoutError(self.timeout_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self._timeout_handler)
        signal.alarm(self.seconds)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)
        if self.suppress and exc_type is TimeoutError:
            return True



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016-2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import warnings
import numpy as np

import librosa
from librosa.util import valid_audio
from librosa.util.exceptions import ParameterError

from pyannote.core import SlidingWindow, SlidingWindowFeature

from soundfile import SoundFile
import soundfile as sf


def get_audio_duration(current_file):
    """Return audio file duration

    Parameters
    ----------
    current_file : dict
        Dictionary given by pyannote.database.

    Returns
    -------
    duration : float
        Audio file duration.
    """

    with SoundFile(current_file["audio"], "r") as f:
        duration = float(f.frames) / f.samplerate

    return duration


def get_audio_sample_rate(current_file):
    """Return audio file sampling rate

    Parameters
    ----------
    current_file : dict
        Dictionary given by pyannote.database.

    Returns
    -------
    sample_rate : int
        Sampling rate
    """
    with SoundFile(current_file["audio"], "r") as f:
        sample_rate = f.samplerate

    return sample_rate


def read_audio(current_file, sample_rate=None, mono=True):
    """Read audio file

    Parameters
    ----------
    current_file : dict
        Dictionary given by pyannote.database.
    sample_rate: int, optional
        Target sampling rate. Defaults to using native sampling rate.
    mono : int, optional
        Convert multi-channel to mono. Defaults to True.

    Returns
    -------
    y : (n_samples, n_channels) np.array
        Audio samples.
    sample_rate : int
        Sampling rate.

    Notes
    -----
    In case `current_file` contains a `channel` key, data of this (1-indexed)
    channel will be returned.

    """

    y, file_sample_rate = sf.read(
        current_file["audio"], dtype="float32", always_2d=True
    )

    # extract specific channel if requested
    channel = current_file.get("channel", None)
    if channel is not None:
        y = y[:, channel - 1 : channel]

    # convert to mono
    if mono and y.shape[1] > 1:
        y = np.mean(y, axis=1, keepdims=True)

    # resample if sample rates mismatch
    if (sample_rate is not None) and (file_sample_rate != sample_rate):
        if y.shape[1] == 1:
            # librosa expects mono audio to be of shape (n,), but we have (n, 1).
            y = librosa.core.resample(y[:, 0], file_sample_rate, sample_rate)[:, None]
        else:
            y = librosa.core.resample(y.T, file_sample_rate, sample_rate).T
    else:
        sample_rate = file_sample_rate

    return y, sample_rate


class RawAudio:
    """Raw audio with on-the-fly data augmentation

    Parameters
    ----------
    sample_rate: int, optional
        Target sampling rate. Defaults to using native sampling rate.
    mono : int, optional
        Convert multi-channel to mono. Defaults to True.
    augmentation : `pyannote.audio.augmentation.Augmentation`, optional
        Data augmentation.
    """

    def __init__(self, sample_rate=None, mono=True, augmentation=None):

        super().__init__()
        self.sample_rate = sample_rate
        self.mono = mono

        self.augmentation = augmentation

        if sample_rate is not None:
            self.sliding_window_ = SlidingWindow(
                start=-0.5 / sample_rate,
                duration=1.0 / sample_rate,
                step=1.0 / sample_rate,
            )

    @property
    def dimension(self):
        return 1

    @property
    def sliding_window(self):
        return self.sliding_window_

    def get_features(self, y, sample_rate):

        # convert to mono
        if self.mono:
            y = np.mean(y, axis=1, keepdims=True)

        # resample if sample rates mismatch
        if (self.sample_rate is not None) and (self.sample_rate != sample_rate):
            if y.shape[1] == 1:
                # librosa expects mono audio to be of shape (n,), but we have (n, 1).
                y = librosa.core.resample(y[:, 0], sample_rate, self.sample_rate)[:, None]
            else:
                y = librosa.core.resample(y.T, sample_rate, self.sample_rate).T
            sample_rate = self.sample_rate

        # augment data
        if self.augmentation is not None:
            y = self.augmentation(y, sample_rate)

        # TODO: how time consuming is this thing (needs profiling...)
        try:
            valid = valid_audio(y[:, 0], mono=True)
        except ParameterError as e:
            msg = f"Something went wrong when augmenting waveform."
            raise ValueError(msg)

        return y

    def __call__(self, current_file, return_sr=False):
        """Obtain waveform

        Parameters
        ----------
        current_file : dict
            `pyannote.database` files.
        return_sr : `bool`, optional
            Return sample rate. Defaults to False

        Returns
        -------
        waveform : `pyannote.core.SlidingWindowFeature`
            Waveform
        sample_rate : `int`
            Only when `return_sr` is set to True
        """

        if "waveform" in current_file:

            if self.sample_rate is None:
                msg = (
                    "`RawAudio` needs to be instantiated with an actual "
                    "`sample_rate` if one wants to use precomputed "
                    "waveform."
                )
                raise ValueError(msg)
            sample_rate = self.sample_rate

            y = current_file["waveform"]

            if len(y.shape) != 2:
                msg = (
                    f"Precomputed waveform should be provided as a "
                    f"(n_samples, n_channels) `np.ndarray`."
                )
                raise ValueError(msg)

        else:
            y, sample_rate = sf.read(
                current_file["audio"], dtype="float32", always_2d=True
            )

        # extract specific channel if requested
        channel = current_file.get("channel", None)
        if channel is not None:
            y = y[:, channel - 1 : channel]

        y = self.get_features(y, sample_rate)

        sliding_window = SlidingWindow(
            start=-0.5 / sample_rate, duration=1.0 / sample_rate, step=1.0 / sample_rate
        )

        if return_sr:
            return (
                SlidingWindowFeature(y, sliding_window),
                sample_rate if self.sample_rate is None else self.sample_rate,
            )

        return SlidingWindowFeature(y, sliding_window)

    def get_context_duration(self):
        return 0.0

    def crop(self, current_file, segment, mode="center", fixed=None):
        """Fast version of self(current_file).crop(segment, **kwargs)

        Parameters
        ----------
        current_file : dict
            `pyannote.database` file.
        segment : `pyannote.core.Segment`
            Segment from which to extract features.
        mode : {'loose', 'strict', 'center'}, optional
            In 'strict' mode, only frames fully included in 'segment' are
            returned. In 'loose' mode, any intersecting frames are returned. In
            'center' mode, first and last frames are chosen to be the ones
            whose centers are the closest to 'focus' start and end times.
            Defaults to 'center'.
        fixed : float, optional
            Overrides `Segment` 'focus' duration and ensures that the number of
            returned frames is fixed (which might otherwise not be the case
            because of rounding errors). Has no effect in 'strict' or 'loose'
            modes.

        Returns
        -------
        waveform : (n_samples, n_channels) numpy array
            Waveform

        See also
        --------
        `pyannote.core.SlidingWindowFeature.crop`
        """

        if self.sample_rate is None:
            msg = (
                "`RawAudio` needs to be instantiated with an actual "
                "`sample_rate` if one wants to use `crop`."
            )
            raise ValueError(msg)

        # find the start and end positions of the required segment
        ((start, end),) = self.sliding_window_.crop(
            segment, mode=mode, fixed=fixed, return_ranges=True
        )

        # this is expected number of samples.
        # this will be useful later in case of on-the-fly resampling
        n_samples = end - start

        if "waveform" in current_file:

            y = current_file["waveform"]

            if len(y.shape) != 2:
                msg = (
                    f"Precomputed waveform should be provided as a "
                    f"(n_samples, n_channels) `np.ndarray`."
                )
                raise ValueError(msg)

            sample_rate = self.sample_rate
            data = y[start:end]

        else:
            # read file with SoundFile, which supports various fomats
            # including NIST sphere
            with SoundFile(current_file["audio"], "r") as audio_file:

                sample_rate = audio_file.samplerate

                # if the sample rates are mismatched,
                # recompute the start and end
                if sample_rate != self.sample_rate:

                    sliding_window = SlidingWindow(
                        start=-0.5 / sample_rate,
                        duration=1.0 / sample_rate,
                        step=1.0 / sample_rate,
                    )
                    ((start, end),) = sliding_window.crop(
                        segment, mode=mode, fixed=fixed, return_ranges=True
                    )

                try:
                    audio_file.seek(start)
                    data = audio_file.read(end - start, dtype="float32", always_2d=True)
                except RuntimeError as e:
                    msg = (
                        f"SoundFile failed to seek-and-read in "
                        f"{current_file['audio']}: loading the whole file..."
                    )
                    warnings.warn(msg)
                    return self(current_file).crop(segment, mode=mode, fixed=fixed)

        # extract specific channel if requested
        channel = current_file.get("channel", None)
        if channel is not None:
            data = data[:, channel - 1 : channel]

        return self.get_features(data, sample_rate)


# # THIS SCRIPT CAN BE USED TO CRASH-TEST THE ON-THE-FLY RESAMPLING

# import numpy as np
# from pyannote.audio.features import RawAudio
# from pyannote.core import Segment
# from pyannote.audio.features.utils import get_audio_duration
# from tqdm import tqdm
#
# TEST_FILE = '/Users/bredin/Corpora/etape/BFMTV_BFMStory_2010-09-03_175900.wav'
# current_file = {'audio': TEST_FILE}
# duration = get_audio_duration(current_file)
#
# for sample_rate in [8000, 16000, 44100, 48000]:
#     raw_audio = RawAudio(sample_rate=sample_rate)
#     for i in tqdm(range(1000), desc=f'{sample_rate:d}Hz'):
#         start = np.random.rand() * (duration - 1.)
#         data = raw_audio.crop(current_file, Segment(start, start + 1), fixed=1.)
#         assert len(data) == sample_rate



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

raise NotImplementedError("FIXME")

from typing import Optional
from pathlib import Path
import yaml

import torch
from functools import partial
from pyannote.core.utils.helper import get_class_by_name

from pyannote.pipeline import Pipeline
from pyannote.pipeline.parameter import Integer
from pyannote.pipeline.parameter import LogUniform
from pyannote.pipeline.parameter import Uniform

from pyannote.core import Annotation
from pyannote.metrics.diarization import GreedyDiarizationErrorRate

from pyannote.audio.labeling.tasks.resegmentation import (
    Resegmentation as _Resegmentation,
)
from pyannote.audio.labeling.tasks.resegmentation import (
    ResegmentationWithOverlap as _ResegmentationWithOverlap,
)


class Resegmentation(Pipeline):
    """Resegmentation pipeline

    Parameters
    ----------
    feature_extraction : `dict`, optional
        Configuration dict for feature extraction.
    architecture : `dict`, optional
        Configuration dict for network architecture.
    overlap : `boolean`, optional
        Assign overlapped speech segments. Defaults to False.
    keep_sad: `boolean`, optional
        Keep speech/non-speech state unchanged. Defaults to False.
    mask : `dict`, optional
        Configuration dict for masking.
        - dimension : `int`, optional
        - log_scale : `bool`, optional
    augmentation : `bool`, optional
        Augment (self-)training data by adding noise from non-speech regions.
        Defaults to False.
    duration : `float`, optional
        Defaults to 2s.
    batch_size : `int`, optional
        Defaults to 32.
    gpu : `boolean`, optional
        Defaults to False.

    Sample configuration file
    -------------------------
    pipeline:
        name: ResegmentationPipeline
        params:
            duration: 3
            batch_size: 32
            gpu: True
            overlap: True
            keep_sad: True
            feature_extraction:
               name: Precomputed
               params:
                  root_dir: /path/to/precomputed/features
            architecture:
               name: StackedLSTM
               params:
                  rnn: LSTM
            mask:
                dimension: 0
                log_scale: True

    preprocessors:
        audio: /path/to/database.yml
        hypothesis:
           name: pyannote.database.util.RTTMLoader
           params:
              train: /path/to/input.train.rttm
              development: /path/to/input.development.rttm
              test: /path/to/input.test.rttm
        overlap:
           name: pyannote.audio.features.Precomputed
           params:
              root_dir: /path/to/precomputed/overlap_scores
        mask:
           name: pyannote.audio.features.Precomputed
           params:
              root_dir: /path/to/precomputed/overlap_scores

    """

    CONFIG_YML = "{experiment_dir}/config.yml"

    # TODO. add support for data augmentation
    def __init__(
        self,
        feature_extraction: Optional[dict] = None,
        architecture: Optional[dict] = None,
        overlap: Optional[bool] = False,
        keep_sad: Optional[bool] = False,
        mask: Optional[dict] = None,
        augmentation: Optional[bool] = False,
        duration: Optional[float] = 2.0,
        batch_size: Optional[float] = 32,
        gpu: Optional[bool] = False,
    ):

        # feature extraction
        if feature_extraction is None:
            from pyannote.audio.features import LibrosaMFCC

            self.feature_extraction_ = LibrosaMFCC(
                e=False,
                De=True,
                DDe=True,
                coefs=19,
                D=True,
                DD=True,
                duration=0.025,
                step=0.010,
                sample_rate=16000,
            )
        else:
            FeatureExtraction = get_class_by_name(
                feature_extraction["name"],
                default_module_name="pyannote.audio.features",
            )
            self.feature_extraction_ = FeatureExtraction(
                **feature_extraction.get("params", {}), augmentation=None
            )

        # network architecture
        if architecture is None:
            from pyannote.audio.models import PyanNet

            self.Architecture_ = PyanNet
            self.architecture_params_ = {"sincnet": {"skip": True}}

        else:
            self.Architecture_ = get_class_by_name(
                architecture["name"], default_module_name="pyannote.audio.models"
            )
            self.architecture_params_ = architecture.get("params", {})

        self.overlap = overlap
        self.keep_sad = keep_sad

        self.mask = mask
        if mask is None:
            self.mask_dimension_ = None
            self.mask_logscale_ = False
        else:
            self.mask_dimension_ = mask["dimension"]
            self.mask_logscale_ = mask["log_scale"]

        self.augmentation = augmentation

        self.duration = duration
        self.batch_size = batch_size
        self.gpu = gpu
        self.device_ = torch.device("cuda") if self.gpu else torch.device("cpu")

        # hyper-parameters
        self.learning_rate = LogUniform(1e-3, 1)
        self.epochs = Integer(10, 50)
        self.ensemble = Integer(1, 5)
        if self.overlap:
            self.overlap_threshold = Uniform(0, 1)

    def initialize(self):
        """Initialize pipeline with current set of parameters"""

        ensemble = min(self.epochs, self.ensemble)

        if self.overlap:
            self._resegmentation = _ResegmentationWithOverlap(
                self.feature_extraction_,
                self.Architecture_,
                self.architecture_params_,
                keep_sad=self.keep_sad,
                mask_dimension=self.mask_dimension_,
                mask_logscale=self.mask_logscale_,
                overlap_threshold=self.overlap_threshold,
                epochs=self.epochs,
                learning_rate=self.learning_rate,
                ensemble=ensemble,
                device=self.device_,
                duration=self.duration,
                batch_size=self.batch_size,
            )

        else:
            self._resegmentation = _Resegmentation(
                self.feature_extraction_,
                self.Architecture_,
                self.architecture_params_,
                keep_sad=self.keep_sad,
                mask_dimension=self.mask_dimension_,
                mask_logscale=self.mask_logscale_,
                epochs=self.epochs,
                learning_rate=self.learning_rate,
                ensemble=ensemble,
                device=self.device_,
                duration=self.duration,
                batch_size=self.batch_size,
            )

    def __call__(self, current_file: dict) -> Annotation:
        """Apply resegmentation

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol. Should contain a
            'hypothesis' key providing diarization before resegmentation (and
            a 'overlap' key in case overlap handling).

        Returns
        -------
        new_hypothesis : `pyannote.core.Annotation`
            Resegmented hypothesis.
        """

        return self._resegmentation.apply(current_file)

    def get_metric(self) -> GreedyDiarizationErrorRate:
        """Return new instance of detection error rate metric"""
        return GreedyDiarizationErrorRate(collar=0.0, skip_overlap=False)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional, Text
from pathlib import Path
from tqdm import tqdm

import torch
from .base import Application
from pyannote.database import get_protocol
from pyannote.database import get_annotated
from pyannote.database import FileFinder
from pyannote.audio.features import Precomputed
from pyannote.audio.features import RawAudio
from pyannote.audio.features.utils import get_audio_duration

from pyannote.database import Subset


class BaseLabeling(Application):
    @property
    def config_default_module(self):
        return "pyannote.audio.labeling.tasks"

    def validate_init(self, protocol_name: Text, subset: Subset = "development"):
        """Initialize validation data

        Parameters
        ----------
        protocol_name : `str`
        subset : {'train', 'development', 'test'}
            Defaults to 'development'.

        Returns
        -------
        validation_data : object
            Validation data.

        """

        preprocessors = self.preprocessors_
        if "audio" not in preprocessors:
            preprocessors["audio"] = FileFinder()
        if "duration" not in preprocessors:
            preprocessors["duration"] = get_audio_duration
        protocol = get_protocol(protocol_name, preprocessors=preprocessors)
        files = getattr(protocol, subset)()

        # convert lazy ProtocolFile to regular dict for multiprocessing
        files = [dict(file) for file in files]

        if isinstance(self.feature_extraction_, (Precomputed, RawAudio)):
            return files

        validation_data = []
        for current_file in tqdm(files, desc="Feature extraction"):
            current_file["features"] = self.feature_extraction_(current_file)
            validation_data.append(current_file)

        return validation_data



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Juan Manuel Coria
# Hervé Bredin - http://herve.niderb.fr

from typing_extensions import Literal
from warnings import warn

import torch
import torch.nn as nn


class TemporalPooling(nn.Module):
    """Pooling strategy over temporal sequences."""

    @staticmethod
    def create(method: Literal["sum", "max", "last", "stats"]) -> nn.Module:
        """Pooling strategy factory. returns an instance of `TemporalPooling` given its name.

        Parameters
        ----------
        method : {'sum', 'max', 'last', 'stats', 'x-vector'}
            Temporal pooling strategy. The `x-vector` method name
            for stats pooling (equivalent to `stats`) is kept for
            retrocompatibility but it will be removed in a future version.
        Returns
        -------
        output : nn.Module
            The temporal pooling strategy object
        """
        if method == "sum":
            klass = SumPool
        elif method == "max":
            klass = MaxPool
        elif method == "last":
            klass = LastPool
        elif method == "stats":
            klass = StatsPool
        elif method == "x-vector":
            klass = StatsPool
            warn(
                "`x-vector` is deprecated and will be removed in a future version. Please use `stats` instead"
            )
        else:
            raise ValueError(f"`{method}` is not a valid temporal pooling method")
        return klass()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("TemporalPooling subclass must implement `forward`")


class SumPool(TemporalPooling):
    """Calculate pooling as the sum over a sequence"""

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        x : `torch.Tensor`, shape (batch_size, seq_len, hidden_size)
            A batch of sequences.

        Returns
        -------
        output : `torch.Tensor`, shape (batch_size, hidden_size)
        """
        return x.sum(dim=1)


class MaxPool(TemporalPooling):
    """Calculate pooling as the maximum over a sequence"""

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        x : `torch.Tensor`, shape (batch_size, seq_len, hidden_size)
            A batch of sequences.

        Returns
        -------
        output : `torch.Tensor`, shape (batch_size, hidden_size)
        """
        return x.max(dim=1)[0]


class LastPool(TemporalPooling):
    """Calculate pooling as the last element of a sequence"""

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        x : `torch.Tensor`, shape (batch_size, seq_len, hidden_size)
            A batch of sequences.

        Returns
        -------
        output : `torch.Tensor`, shape (batch_size, hidden_size)
        """
        return x[:, -1]


class StatsPool(TemporalPooling):
    """Calculate pooling as the concatenated mean and standard deviation of a sequence"""

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Parameters
        ----------
        x : `torch.Tensor`, shape (batch_size, seq_len, hidden_size)
            A batch of sequences.

        Returns
        -------
        output : `torch.Tensor`, shape (batch_size, 2 * hidden_size)
        """
        mean, std = torch.mean(x, dim=1), torch.std(x, dim=1)
        return torch.cat((mean, std), dim=1)


class Pooling(nn.Module):
    """Pooling over the time dimension

    Parameters
    ----------
    method : {"last", "max", "average"}, optional
        Use "max" for max pooling, "average" for average pooling.
        Use "average" for average pooling.
        Use "last" for returning the last element of the sequence.
    bidirectional : bool, optional
        When using "last" pooling, indicate whether the input sequence should
        be considered as the output of a bidirectional recurrent layer, in which
        case the last element in both directions are concatenated.
    """

    def __init__(
        self,
        n_features,
        method: Literal["last", "max", "average"] = None,
        bidirectional: bool = None,
    ):
        super().__init__()

        if method == "last" and bidirectional is None:
            msg = "'last' pooling expects an additional 'bidirectional' parameter."
            raise ValueError(msg)

        self.n_features = n_features
        self.method = method
        self.bidirectional = bidirectional

    def forward(self, sequences: torch.Tensor) -> torch.Tensor:
        """Temporal pooling

        Parameters
        ----------
        sequences : torch.Tensor
            Input sequences with shape (batch_size, n_frames, n_features)

        Returns
        -------
        pooled : torch.Tensor
            Pooled sequences with shape (batch_size, n_features)
        """

        if self.method is None:
            return sequences

        if self.method == "last":
            if self.bidirectional:
                batch_size, n_frames, _ = sequences.shape
                reshaped = sequences.view(batch_size, n_frames, 2, -1)
                return torch.cat([reshaped[:, -1, 0], reshaped[:, 0, 1]], dim=1)
            else:
                return sequences[:, -1]

        if self.method == "max":
            return torch.max(sequences, dim=1, keepdim=False, out=None)[0]

        if self.method == "average":
            return torch.mean(sequences, dim=1, keepdim=False, out=None)

    @property
    def dimension(self):
        "Dimension of output features"
        return self.n_features



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from typing import Union
from typing import List
from typing import Optional

NoiseCollection = Union[str, List[str]]

import random
import numpy as np

from pyannote.core import Segment
from pyannote.core.utils.random import random_subsegment

from pyannote.audio.features import RawAudio
from pyannote.audio.features.utils import get_audio_duration

from pyannote.database import FileFinder
from pyannote.database import get_protocol


normalize = lambda wav: wav / (np.sqrt(np.mean(wav ** 2)) + 1e-8)


class Noise:
    """Noise generator

    Parameters
    ----------
    collection : str or list of str
        `pyannote.database` collection(s) used for adding noise.
        Defaults to "MUSAN.Collection.BackgroundNoise"
    """

    def __init__(self, collection: Optional[NoiseCollection] = None):

        if collection is None:
            collection = "MUSAN.Collection.BackgroundNoise"

        if not isinstance(collection, (list, tuple)):
            collection = [collection]

        self.collection = collection

        self.files_ = []
        preprocessors = {"audio": FileFinder(), "duration": get_audio_duration}
        for collection in self.collection:
            protocol = get_protocol(collection, preprocessors=preprocessors)
            self.files_.extend(protocol.files())

    def __call__(self, n_samples: int, sample_rate: int) -> np.ndarray:
        """Generate noise

        Parameters
        ----------
        n_samples : int
        sample_rate : int

        Returns
        -------
        noise : (n_samples, 1) np.ndarray
        """

        target_duration = n_samples / sample_rate

        raw_audio = RawAudio(sample_rate=sample_rate, mono=True)

        # accumulate enough noise to cover duration
        noises = []
        left = target_duration
        while left > 0:

            # select noise file at random
            file = random.choice(self.files_)
            duration = file["duration"]

            # if noise file is longer than what is needed, crop it
            if duration > left:
                segment = next(random_subsegment(Segment(0, duration), left))
                noise = raw_audio.crop(file, segment, mode="center", fixed=left)
                left = 0

            # otherwise, take the whole file
            else:
                noise = raw_audio(file).data
                left -= duration

            noise = normalize(noise)
            noises.append(noise)

        # concatenate
        noise = np.vstack(noises)
        # TODO: use fade-in between concatenated noises

        return noise



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import torch
import torch.nn as nn

try:
    from typing import Literal
except ImportError as e:
    from typing_extensions import Literal


class Scaling(nn.Module):
    """Scale feature vectors

    Parameters
    ----------
    n_features : int
        Number of input features.
    method : {"unit", "logistic"}, optional
        Defaults to no scaling.
    """

    def __init__(self, n_features: int, method: Literal["fixed", "logistic"] = None):
        super().__init__()
        self.n_features = n_features
        self.method = method

        if self.method == "logistic":
            self.batch_norm = nn.BatchNorm1d(
                1, eps=1e-5, momentum=0.1, affine=True, track_running_stats=True
            )
            self.activation = nn.Sigmoid()

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """Scale features

        Parameters
        ----------
        features : torch.Tensor
            Input features of shape (batch_size, *, n_features)

        Returns
        -------
        scaled : torch.Tensor
            Scaled features of shape (batch_size, *, n_features)
        """

        if self.method is None:
            return features

        norm = features.norm(p=2, dim=-1, keepdim=True)

        if self.method == "unit":
            new_norm = 1.0

        if self.method == "logistic":
            new_norm = self.activation(self.batch_norm(norm))

        return new_norm / (norm + 1e-6) * features

    @property
    def dimension(self):
        return self.n_features



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from pathlib import Path
from typing import Text
from typing import Union
from typing import Dict
from functools import partial
from pyannote.database import ProtocolFile
from pyannote.core import Segment
from pyannote.core import SlidingWindowFeature
import numpy as np

Wrappable = Union[
    "Precomputed", "Pretrained", "RawAudio", "FeatureExtraction", Dict, Text, Path
]

# this needs to go here to make Wrapper instances pickable
def _use_existing_key(key, file):
    return file[key]


class Wrapper:
    """FeatureExtraction-compliant wrapper

    Parameters
    ----------
    wrappable : Wrappable
        Wrappable object. See "Usage" section for a detailed description.
    **params : Dict
        Keyword parameters passed to the wrapped object when supported.

    Usage
    -----
    If `wrappable` already complies with the `FeatureExtraction` API , it is
    kept unchanged. This includes instances of any `FeatureExtraction` subclass,
    `RawAudio` instances, `Precomputed` instances, and `Pretrained instances.
    In this case, keyword parameters are not used.

    * If `wrappable` is a `Path` to a directory containing precomputed features
      or scores (e.g. the one created by `pyannote-audio [...] apply [...]`), it
      stands for:

      Precomputed(root_dir=wrappable, **params)

    * If `wrappable` is a `Path` to the validation directory created by calling
      `pyannote-audio [...] validate [...]`, it stands for:

      Pretrained(validate_dir=wrappable, **params)

    * If `wrappable` is a `Path` to a checkpoint created by calling
      `pyannote-audio [...] train [...]`, (i.e. with the following structure:
      '{root_dir}/train/{protocol}.train/weights/{epoch}.pt)', it stands for:

      Pretrained(validate_dir='{root_dir}/train/{protocol}.train/validate/fake',
                 epoch=int({epoch}), **params)

    * If `wrappable` is a `Text` containing the name of an existing `torch.hub`
      model, it stands for:

      torch.hub.load('pyannote/pyannote-audio', wrappable, **params)

    * If `wrappable` is a `Text` starting with '@' such as '@key', it stands for:

      lambda current_file: current_file['key']

    In any other situation, it will raise an error.

    Notes
    -----
    It is also possible to provide a `Dict` `wrappable`, in which case it is
    expected to contain a unique key which is the name of a `torch.hub` model
    (or any supported `Path` described above), whose corresponding value is a
    dictionary of custom parameters. For instance,

      Wrapper({'sad': {'step': 0.1}}) is the same as Wrapper('sad', step=0.1)

    This is especially useful in `pyannote-pipeline` configuration files:

        ~~~~~ content of 'config.yml' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
          pipeline:
             name: pyannote.audio.pipeline.SpeechActivityDetection
             params:
                scores:
                   sad:
                      step: 0.1
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    """

    def __init__(self, wrappable: Wrappable, **params):
        super().__init__()

        from pyannote.audio.features import Pretrained
        from pyannote.audio.features import Precomputed
        from pyannote.audio.features import FeatureExtraction
        from pyannote.audio.features import RawAudio

        scorer = None
        msg = ""

        # corner
        if isinstance(wrappable, dict):
            wrappable, custom_params = dict(wrappable).popitem()
            params.update(**custom_params)

        # If `wrappable` already complies with the `FeatureExtraction` API , it
        # is kept unchanged. This includes instances of any `FeatureExtraction`
        # subclass,`RawAudio` instances, `Precomputed` instances, and
        # `Pretrained` instances.
        if isinstance(
            wrappable, (FeatureExtraction, RawAudio, Pretrained, Precomputed)
        ):
            scorer = wrappable

        elif Path(wrappable).is_dir():
            directory = Path(wrappable)

            # If `wrappable` is a `Path` to a directory containing precomputed
            # features or scores, wrap the corresponding `Precomputed` instance
            try:
                scorer = Precomputed(root_dir=directory)
            except Exception as e:
                scorer = None

            # If `wrappable` is a `Path` to a validation directory,
            # wrap the corresponding `Pretrained` instance
            if scorer is None:
                try:
                    scorer = Pretrained(validate_dir=directory, **params)
                except Exception as e:
                    scorer = None

            if scorer is None:
                msg = (
                    f'"{wrappable}" directory does not seem to be the path '
                    f"to precomputed features nor the path to a model "
                    f"validation step."
                )

        # If `wrappable` is a `Path` to a pretrined model checkpoint,
        # wrap the corresponding `Pretrained` instance
        elif Path(wrappable).is_file():
            checkpoint = Path(wrappable)

            try:
                validate_dir = checkpoint.parents[1] / "validate" / "fake"
                epoch = int(checkpoint.stem)
                scorer = Pretrained(validate_dir=validate_dir, epoch=epoch, **params)
            except Exception as e:
                msg = (
                    f'"{wrappable}" directory does not seem to be the path '
                    f"to a pretrained model checkpoint."
                )
                scorer = None

        elif isinstance(wrappable, Text):

            # If `wrappable` is a `Text` starting with '@' such as '@key',
            # it means that one should read the "key" key of protocol files
            if wrappable.startswith("@"):
                key = wrappable[1:]

                scorer = partial(_use_existing_key, key)
                # scorer = lambda current_file: current_file[key]

            # If `wrappable` is a `Text` containing the name of an existing
            # `torch.hub` model, wrap the corresponding `Pretrained`.
            else:
                try:
                    import torch

                    scorer = torch.hub.load(
                        "pyannote/pyannote-audio", wrappable, **params
                    )
                    if not isinstance(scorer, Pretrained):
                        msg = (
                            f'"{wrappable}" exists on torch.hub but does not '
                            f"return a `Pretrained` model instance."
                        )
                        scorer = None

                except Exception as e:
                    msg = (
                        f"Could not load {wrappable} model from torch.hub. "
                        f"The following exception was raised:\n{e}"
                    )
                    scorer = None

        # warn the user the something went wrong
        if scorer is None:
            raise ValueError(msg)

        self.scorer_ = scorer

    def crop(
        self,
        current_file: ProtocolFile,
        segment: Segment,
        mode: Text = "center",
        fixed: float = None,
    ) -> np.ndarray:
        """Extract frames from a specific region

        Parameters
        ----------
        current_file : ProtocolFile
            Protocol file
        segment : Segment
            Region of the file to process.
        mode : {'loose', 'strict', 'center'}, optional
            In 'strict' mode, only frames fully included in 'segment' support are
            returned. In 'loose' mode, any intersecting frames are returned. In
            'center' mode, first and last frames are chosen to be the ones
            whose centers are the closest to 'segment' start and end times.
            Defaults to 'center'.
        fixed : float, optional
            Overrides 'segment' duration and ensures that the number of
            returned frames is fixed (which might otherwise not be the case
            because of rounding errors).

        Returns
        -------
        frames : np.ndarray
            Frames.
        """

        from pyannote.audio.features import Precomputed
        from pyannote.audio.features import Pretrained
        from pyannote.audio.features import RawAudio
        from pyannote.audio.features import FeatureExtraction

        if isinstance(
            self.scorer_, (FeatureExtraction, RawAudio, Pretrained, Precomputed)
        ):
            return self.scorer_.crop(current_file, segment, mode=mode, fixed=fixed)

        return self.scorer_(current_file).crop(
            segment, mode=mode, fixed=fixed, return_data=True
        )

    def __call__(self, current_file) -> SlidingWindowFeature:
        """Extract frames from the whole file

        Parameters
        ----------
        current_file : ProtocolFile
            Protocol file

        Returns
        -------
        frames : np.ndarray
            Frames.
        """
        return self.scorer_(current_file)

    # used to "inherit" most scorer_ attributes
    def __getattr__(self, name):

        # prevents a "RecursionError: maximum recursion depth exceeded" when pickling Wrapper
        # https://stackoverflow.com/questions/49380224/how-to-make-classes-with-getattr-pickable
        if "scorer_" not in vars(self):
            raise AttributeError

        return getattr(self.scorer_, name)

    def __setattr__(self, name, value):
        if name == "scorer_":
            object.__setattr__(self, name, value)

        else:
            setattr(self.scorer_, name, value)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

# TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
# Update VGGVox to latest API. Commenting it out for now.
# TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO

# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# from .utils import get_conv2d_output_shape


# class VGGVox(nn.Module):
#     """VGGVox implementation
#
#     Reference
#     ---------
#     Arsha Nagrani, Joon Son Chung, Andrew Zisserman. "VoxCeleb: a large-scale
#     speaker identification dataset."
#
#     """
#
#     def __init__(self, specifications, dimension=256):
#
#         super().__init__()
#         self.specifications = specifications
#         self.n_features_ = specifications['X']['dimension']
#
#         if self.n_features_ < 97:
#             msg = (f'VGGVox expects features with at least 97 dimensions '
#                    f'(here, n_features = {self.n_features_:d})')
#             raise ValueError(msg)
#
#         self.dimension = dimension
#
#         h = self.n_features_  # 512 in VoxCeleb paper. 201 in practice.
#         w = 301 # typically 3s with 10ms steps
#
#         self.conv1_ = nn.Conv2d(1, 96, (7, 7), stride=(2, 2), padding=1)
#         # 254 x 148 when n_features = 512
#         # 99 x 148 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (7, 7), stride=(2, 2), padding=1)
#
#         self.bn1_ = nn.BatchNorm2d(96)
#         self.mpool1_ = nn.MaxPool2d((3, 3), stride=(2, 2))
#         # 126 x 73 when n_features = 512
#         # 49 x 73 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (3, 3), stride=(2, 2))
#
#         self.conv2_ = nn.Conv2d(96, 256, (5, 5), stride=(2, 2), padding=1)
#         # 62 x 36 when n_features = 512
#         # 24 x 36 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (5, 5), stride=(2, 2), padding=1)
#
#         self.bn2_ = nn.BatchNorm2d(256)
#         self.mpool2_ = nn.MaxPool2d((3, 3), stride=(2, 2))
#         # 30 x 17 when n_features = 512
#         # 11 x 17 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (3, 3), stride=(2, 2))
#
#         self.conv3_ = nn.Conv2d(256, 256, (3, 3), stride=(1, 1), padding=1)
#         # 30 x 17 when n_features = 512
#         # 11 x 17 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (3, 3), stride=(1, 1), padding=1)
#
#         self.bn3_ = nn.BatchNorm2d(256)
#
#         self.conv4_ = nn.Conv2d(256, 256, (3, 3), stride=(1, 1), padding=1)
#         # 30 x 17 when n_features = 512
#         # 11 x 17 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (3, 3), stride=(1, 1), padding=1)
#
#         self.bn4_ = nn.BatchNorm2d(256)
#
#         self.conv5_ = nn.Conv2d(256, 256, (3, 3), stride=(1, 1), padding=1)
#         # 30 x 17 when n_features = 512
#         # 11 x 17 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (3, 3), stride=(1, 1), padding=1)
#
#         self.bn5_ = nn.BatchNorm2d(256)
#
#         self.mpool5_ = nn.MaxPool2d((5, 3), stride=(3, 2))
#         # 9 x 8 when n_features = 512
#         # 3 x 8 when n_features = 201
#         h, w = get_conv2d_output_shape((h, w), (5, 3), stride=(3, 2))
#
#         self.fc6_ = nn.Conv2d(256, 4096, (h, 1), stride=(1, 1))
#         # 1 x 8
#         h, w = get_conv2d_output_shape((h, w), (h, 1), stride=(1, 1))
#
#         self.fc7_ = nn.Linear(4096, 1024)
#         self.fc8_ = nn.Linear(1024, self.dimension)
#
#
#     def forward(self, sequences):
#         """Embed sequences
#
#         Parameters
#         ----------
#         sequences : torch.Tensor (batch_size, n_samples, n_features)
#             Batch of sequences.
#
#         Returns
#         -------
#         embeddings : torch.Tensor (batch_size, dimension)
#             Batch of embeddings.
#         """
#
#         # reshape batch to (batch_size, n_channels, n_features, n_samples)
#         batch_size, n_samples, n_features = sequences.size()
#
#         if n_features != self.n_features_:
#             msg = (f'Mismatch in feature dimension '
#                    f'(should be: {self.n_features_:d}, is: {n_features:d})')
#             raise ValueError(msg)
#
#         if n_samples < 65:
#             msg = (f'VGGVox expects sequences with at least 65 samples '
#                    f'(here, n_samples = {n_samples:d})')
#             raise ValueError(msg)
#
#         x = torch.transpose(sequences, 1, 2).view(
#             batch_size, 1, n_features, n_samples)
#
#         # conv1. shape => 254 x 148 => 126 x 73
#         x = self.mpool1_(F.relu(self.bn1_(self.conv1_(x))))
#
#         # conv2. shape =>
#         x = self.mpool2_(F.relu(self.bn2_(self.conv2_(x))))
#
#         # conv3. shape = 62 x 36
#         x = F.relu(self.bn3_(self.conv3_(x)))
#
#         # conv4. shape = 30 x 17
#         x = F.relu(self.bn4_(self.conv4_(x)))
#
#         # conv5. shape = 30 x 17
#         x = self.mpool5_(F.relu(self.bn5_(self.conv5_(x))))
#
#         # fc6. shape =
#         x = F.dropout(F.relu(self.fc6_(x)))
#
#         # (average) temporal pooling. shape =
#         x = torch.mean(x, dim=-1)
#
#         # fc7. shape =
#         x = x.view(x.size(0), -1)
#         x = F.dropout(F.relu(self.fc7_(x)))
#
#         # fc8. shape =
#         x = self.fc8_(x)
#
#         return x



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


class Callback:
    def on_train_start(self, trainer):
        pass

    def on_epoch_start(self, trainer):
        pass

    def on_batch_start(self, trainer, batch):
        return batch

    def on_batch_end(self, trainer, loss):
        """Called after parameters update

        Parameters
        ----------
        trainer : `Trainer`
        loss : `dict`
        """
        pass

    def on_epoch_end(self, trainer):
        pass

    def on_train_end(self, trainer):
        pass


class Debug(Callback):
    def on_train_start(self, trainer):
        print("")
        print("-- Debug callback -----------------")
        print("on_train_start")
        print("-----------------------------------")
        print("")

    def on_epoch_start(self, trainer):
        print("")
        print("-- Debug callback -----------------")
        print("on_epoch_start")
        print("-----------------------------------")
        print("")

    def on_batch_start(self, trainer, batch):
        print("")
        print("-- Debug callback -----------------")
        print("on_batch_start")
        print("-----------------------------------")
        print("")
        return batch

    def on_batch_end(self, trainer, loss):
        print("")
        print("-- Debug callback -----------------")
        print("on_batch_end")
        print("-----------------------------------")
        print("")

    def on_epoch_end(self, trainer):
        print("")
        print("-- Debug callback -----------------")
        print("on_epoch_end")
        print("-----------------------------------")
        print("")

    def on_train_end(self, trainer):
        print("")
        print("-- Debug callback -----------------")
        print("on_train_end")
        print("-----------------------------------")
        print("")


class Callbacks:
    def __init__(self, callbacks):
        super().__init__()
        self.callbacks = callbacks

    def on_train_start(self, trainer):
        for callback in self.callbacks:
            callback.on_train_start(trainer)

    def on_epoch_start(self, trainer):
        for callback in self.callbacks:
            callback.on_epoch_start(trainer)

    def on_batch_start(self, trainer, batch):
        for callback in self.callbacks:
            batch = callback.on_batch_start(trainer, batch)
        return batch

    def on_batch_end(self, trainer, loss):
        """Called after parameters update

        Parameters
        ----------
        trainer : `Trainer`
        loss : `dict`
        """
        trainer.on_batch_end(loss)
        for callback in self.callbacks:
            callback.on_batch_end(trainer, loss)

    def on_epoch_end(self, trainer):
        trainer.on_epoch_end()
        for callback in self.callbacks:
            callback.on_epoch_end(trainer)

    def on_train_end(self, trainer):
        trainer.on_train_end()
        for callback in self.callbacks:
            callback.on_train_end(trainer)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr
# Juan Manuel Coria

from typing import Optional
from typing import Text

import torch
import torch.nn as nn

from .sincnet import SincNet
from .tdnn import XVectorNet
from .pooling import TemporalPooling


from .convolutional import Convolutional
from .recurrent import Recurrent
from .linear import Linear
from .pooling import Pooling
from .scaling import Scaling

from pyannote.audio.train.model import Model
from pyannote.audio.train.model import Resolution
from pyannote.audio.train.model import RESOLUTION_CHUNK
from pyannote.audio.train.model import RESOLUTION_FRAME
from pyannote.audio.train.task import Task


class RNN(nn.Module):
    """Recurrent layers

    Parameters
    ----------
    n_features : `int`
        Input feature shape.
    unit : {'LSTM', 'GRU'}, optional
        Defaults to 'LSTM'.
    hidden_size : `int`, optional
        Number of features in the hidden state h. Defaults to 16.
    num_layers : `int`, optional
        Number of recurrent layers. Defaults to 1.
    bias : `boolean`, optional
        If False, then the layer does not use bias weights. Defaults to True.
    dropout : `float`, optional
        If non-zero, introduces a Dropout layer on the outputs of each layer
        except the last layer, with dropout probability equal to dropout.
        Defaults to 0.
    bidirectional : `boolean`, optional
        If True, becomes a bidirectional RNN. Defaults to False.
    concatenate : `boolean`, optional
        Concatenate output of each layer instead of using only the last one
        (which is the default behavior).
    pool : {'sum', 'max', 'last', 'x-vector'}, optional
        Temporal pooling strategy. Defaults to no pooling.
    """

    def __init__(
        self,
        n_features,
        unit="LSTM",
        hidden_size=16,
        num_layers=1,
        bias=True,
        dropout=0,
        bidirectional=False,
        concatenate=False,
        pool=None,
    ):
        super().__init__()

        self.n_features = n_features

        self.unit = unit
        Klass = getattr(nn, self.unit)

        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.concatenate = concatenate
        self.pool = pool
        self.pool_ = TemporalPooling.create(pool) if pool is not None else None

        if num_layers < 1:
            msg = '"bidirectional" must be set to False when num_layers < 1'
            if bidirectional:
                raise ValueError(msg)
            msg = '"concatenate" must be set to False when num_layers < 1'
            if concatenate:
                raise ValueError(msg)
            return

        if self.concatenate:

            self.rnn_ = nn.ModuleList([])
            for i in range(self.num_layers):

                if i > 0:
                    input_dim = self.hidden_size
                    if self.bidirectional:
                        input_dim *= 2
                else:
                    input_dim = self.n_features

                if i + 1 == self.num_layers:
                    dropout = 0
                else:
                    dropout = self.dropout

                rnn = Klass(
                    input_dim,
                    self.hidden_size,
                    num_layers=1,
                    bias=self.bias,
                    batch_first=True,
                    dropout=dropout,
                    bidirectional=self.bidirectional,
                )

                self.rnn_.append(rnn)

        else:
            self.rnn_ = Klass(
                self.n_features,
                self.hidden_size,
                num_layers=self.num_layers,
                bias=self.bias,
                batch_first=True,
                dropout=self.dropout,
                bidirectional=self.bidirectional,
            )

    def forward(self, features, return_intermediate=False):
        """Apply recurrent layer (and optional temporal pooling)

        Parameters
        ----------
        features : `torch.Tensor`
            Features shaped as (batch_size, n_frames, n_features)
        return_intermediate : `boolean`, optional
            Return intermediate RNN hidden state.

        Returns
        -------
        output : `torch.Tensor`
            TODO. Shape depends on parameters...
        intermediate : `torch.Tensor`
            (num_layers, batch_size, hidden_size * num_directions)
        """

        if self.num_layers < 1:

            if return_intermediate:
                msg = (
                    '"return_intermediate" must be set to False ' "when num_layers < 1"
                )
                raise ValueError(msg)

            output = features

        else:

            if return_intermediate:
                num_directions = 2 if self.bidirectional else 1

            if self.concatenate:

                if return_intermediate:
                    msg = (
                        '"return_intermediate" is not supported '
                        'when "concatenate" is True'
                    )
                    raise NotADirectoryError(msg)

                outputs = []

                hidden = None
                output = None
                # apply each layer separately...
                for i, rnn in enumerate(self.rnn_):
                    if i > 0:
                        output, hidden = rnn(output, hidden)
                    else:
                        output, hidden = rnn(features)
                    outputs.append(output)

                # ... and concatenate their output
                output = torch.cat(outputs, dim=2)

            else:
                output, hidden = self.rnn_(features)

                if return_intermediate:
                    if self.unit == "LSTM":
                        h = hidden[0]
                    elif self.unit == "GRU":
                        h = hidden

                    # to (num_layers, batch_size, num_directions * hidden_size)
                    h = h.view(self.num_layers, num_directions, -1, self.hidden_size)
                    intermediate = (
                        h.transpose(2, 1)
                        .contiguous()
                        .view(self.num_layers, -1, num_directions * self.hidden_size)
                    )

        if self.pool_ is not None:
            output = self.pool_(output)

        if return_intermediate:
            return output, intermediate

        return output

    def dimension():
        doc = "Output features dimension."

        def fget(self):
            if self.num_layers < 1:
                dimension = self.n_features
            else:
                dimension = self.hidden_size

            if self.bidirectional:
                dimension *= 2

            if self.concatenate:
                dimension *= self.num_layers

            if self.pool == "x-vector":
                dimension *= 2

            return dimension

        return locals()

    dimension = property(**dimension())

    def intermediate_dimension(self, layer):
        if self.num_layers < 1:
            dimension = self.n_features
        else:
            dimension = self.hidden_size

        if self.bidirectional:
            dimension *= 2

        return dimension


class FF(nn.Module):
    """Feedforward layers

    Parameters
    ----------
    n_features : `int`
        Input dimension.
    hidden_size : `list` of `int`, optional
        Linear layers hidden dimensions. Defaults to [16, ].
    """

    def __init__(self, n_features, hidden_size=[16,]):
        super().__init__()

        self.n_features = n_features
        self.hidden_size = hidden_size

        self.linear_ = nn.ModuleList([])
        for hidden_size in self.hidden_size:
            linear = nn.Linear(n_features, hidden_size, bias=True)
            self.linear_.append(linear)
            n_features = hidden_size

    def forward(self, features):
        """

        Parameters
        ----------
        features : `torch.Tensor`
            (batch_size, n_samples, n_features) or (batch_size, n_features)

        Returns
        -------
        output : `torch.Tensor`
            (batch_size, n_samples, hidden_size[-1]) or (batch_size, hidden_size[-1])
        """

        output = features
        for linear in self.linear_:
            output = linear(output)
            output = torch.tanh(output)
        return output

    def dimension():
        doc = "Output dimension."

        def fget(self):
            if self.hidden_size:
                return self.hidden_size[-1]
            return self.n_features

        return locals()

    dimension = property(**dimension())


class Embedding(nn.Module):
    """Embedding

    Parameters
    ----------
    n_features : `int`
        Input dimension.
    batch_normalize : `boolean`, optional
        Apply batch normalization. This is more or less equivalent to
        embedding whitening.
    scale : {"fixed", "logistic"}, optional
        Scaling method. Defaults to no scaling.
    unit_normalize : deprecated in favor of 'scale'
    """

    def __init__(
        self,
        n_features: int,
        batch_normalize: bool = False,
        scale: Text = None,
        unit_normalize: bool = False,
    ):
        super().__init__()

        self.n_features = n_features

        self.batch_normalize = batch_normalize
        if self.batch_normalize:
            self.batch_normalize_ = nn.BatchNorm1d(
                n_features, eps=1e-5, momentum=0.1, affine=False
            )

        self.scale = scale
        self.scaling = Scaling(n_features, method=scale)

        if unit_normalize is True:
            msg = f"'unit_normalize' has been deprecated in favor of 'scale'."
            raise ValueError(msg)

    def forward(self, embedding):

        if self.batch_normalize:
            embedding = self.batch_normalize_(embedding)

        return self.scaling(embedding)

    @property
    def dimension(self):
        """Output dimension."""
        return self.n_features


class PyanNet(Model):
    """waveform -> SincNet -> RNN [-> merge] [-> time_pool] -> FC -> output

    Parameters
    ----------
    sincnet : `dict`, optional
        SincNet parameters. Defaults to `pyannote.audio.models.sincnet.SincNet`
        default parameters. Use {'skip': True} to use handcrafted features
        instead of waveforms: [ waveform -> SincNet -> RNN -> ... ] then
        becomes [ features -> RNN -> ...].
    rnn : `dict`, optional
        Recurrent network parameters. Defaults to `RNN` default parameters.
    ff : `dict`, optional
        Feed-forward layers parameters. Defaults to `FF` default parameters.
    embedding : `dict`, optional
        Embedding parameters. Defaults to `Embedding` default parameters. This
        only has effect when model is used for representation learning.
    """

    @staticmethod
    def get_alignment(task: Task, sincnet=None, **kwargs):
        """Get frame alignment"""

        if sincnet is None:
            sincnet = dict()

        if sincnet.get("skip", False):
            return "center"

        return SincNet.get_alignment(task, **sincnet)

    @staticmethod
    def get_resolution(
        task: Task,
        sincnet: Optional[dict] = None,
        rnn: Optional[dict] = None,
        **kwargs,
    ) -> Resolution:
        """Get sliding window used for feature extraction

        Parameters
        ----------
        task : Task
        sincnet : dict, optional
        rnn : dict, optional

        Returns
        -------
        sliding_window : `pyannote.core.SlidingWindow` or {`window`, `frame`}
            Returns RESOLUTION_CHUNK if model returns one vector per input
            chunk, RESOLUTION_FRAME if model returns one vector per input
            frame, and specific sliding window otherwise.
        """

        if rnn is None:
            rnn = {"pool": None}

        if rnn.get("pool", None) is not None:
            return RESOLUTION_CHUNK

        if sincnet is None:
            sincnet = {"skip": False}

        if sincnet.get("skip", False):
            return RESOLUTION_FRAME

        return SincNet.get_resolution(task, **sincnet)

    def init(
        self,
        sincnet: Optional[dict] = None,
        rnn: Optional[dict] = None,
        ff: Optional[dict] = None,
        embedding: Optional[dict] = None,
    ):
        """waveform -> SincNet -> RNN [-> merge] [-> time_pool] -> FC -> output

        Parameters
        ----------
        sincnet : `dict`, optional
            SincNet parameters. Defaults to `pyannote.audio.models.sincnet.SincNet`
            default parameters. Use {'skip': True} to use handcrafted features
            instead of waveforms: [ waveform -> SincNet -> RNN -> ... ] then
            becomes [ features -> RNN -> ...].
        rnn : `dict`, optional
            Recurrent network parameters. Defaults to `RNN` default parameters.
        ff : `dict`, optional
            Feed-forward layers parameters. Defaults to `FF` default parameters.
        embedding : `dict`, optional
            Embedding parameters. Defaults to `Embedding` default parameters. This
            only has effect when model is used for representation learning.
        """

        n_features = self.n_features

        if sincnet is None:
            sincnet = dict()
        self.sincnet = sincnet

        if not sincnet.get("skip", False):
            if n_features != 1:
                msg = (
                    f"SincNet only supports mono waveforms. "
                    f"Here, waveform has {n_features} channels."
                )
                raise ValueError(msg)
            self.sincnet_ = SincNet(**sincnet)
            n_features = self.sincnet_.dimension

        if rnn is None:
            rnn = dict()
        self.rnn = rnn
        self.rnn_ = RNN(n_features, **rnn)
        n_features = self.rnn_.dimension

        if ff is None:
            ff = dict()
        self.ff = ff
        self.ff_ = FF(n_features, **ff)
        n_features = self.ff_.dimension

        if self.task.is_representation_learning:
            if embedding is None:
                embedding = dict()
            self.embedding = embedding
            self.embedding_ = Embedding(n_features, **embedding)
            return

        self.linear_ = nn.Linear(n_features, len(self.classes), bias=True)
        self.activation_ = self.task.default_activation

    def forward(self, waveforms, return_intermediate=None):
        """Forward pass

        Parameters
        ----------
        waveforms : (batch_size, n_samples, 1) `torch.Tensor`
            Batch of waveforms. In case SincNet is skipped, a tensor with shape
            (batch_size, n_samples, n_features) is expected.
        return_intermediate : `int`, optional
            Index of RNN layer. Returns RNN intermediate hidden state.
            Defaults to only return the final output.

        Returns
        -------
        output : `torch.Tensor`
            Final network output.
        intermediate : `torch.Tensor`
            Intermediate network output (only when `return_intermediate`
            is provided).
        """

        if self.sincnet.get("skip", False):
            output = waveforms
        else:
            output = self.sincnet_(waveforms)

        if return_intermediate is None:
            output = self.rnn_(output)
        else:
            if return_intermediate == 0:
                intermediate = output
                output = self.rnn_(output)
            else:
                return_intermediate -= 1
                # get RNN final AND intermediate outputs
                output, intermediate = self.rnn_(output, return_intermediate=True)
                # only keep hidden state of requested layer
                intermediate = intermediate[return_intermediate]

        output = self.ff_(output)

        if self.task.is_representation_learning:
            return self.embedding_(output)

        output = self.linear_(output)
        output = self.activation_(output)

        if return_intermediate is None:
            return output
        return output, intermediate

    @property
    def dimension(self):
        if self.task.is_representation_learning:
            return self.embedding_.dimension

        return Model.dimension.fget(self)

    def intermediate_dimension(self, layer):
        if layer == 0:
            return self.sincnet_.dimension
        return self.rnn_.intermediate_dimension(layer - 1)


class SincTDNN(Model):
    """waveform -> SincNet -> XVectorNet (TDNN -> FC) -> output

    Parameters
    ----------
    sincnet : `dict`, optional
        SincNet parameters. Defaults to `pyannote.audio.models.sincnet.SincNet`
        default parameters.
    tdnn : `dict`, optional
        X-Vector Time-Delay neural network parameters.
        Defaults to `pyannote.audio.models.tdnn.XVectorNet` default parameters.
    embedding : `dict`, optional
        Embedding parameters. Defaults to `Embedding` default parameters. This
        only has effect when model is used for representation learning.
    """

    @staticmethod
    def get_alignment(task: Task, sincnet=None, **kwargs):
        """Get frame alignment"""

        if sincnet is None:
            sincnet = dict()

        return SincNet.get_alignment(task, **sincnet)

    supports_packed = False

    @staticmethod
    def get_resolution(
        task: Task, sincnet: Optional[dict] = None, **kwargs
    ) -> Resolution:
        """Get sliding window used for feature extraction

        Parameters
        ----------
        task : Task
        sincnet : dict, optional

        Returns
        -------
        sliding_window : `pyannote.core.SlidingWindow` or {`window`, `frame`}
        """

        # TODO add support for frame-wise and sequence labeling tasks
        # TODO https://github.com/pyannote/pyannote-audio/issues/290
        return RESOLUTION_CHUNK

    def init(
        self,
        sincnet: Optional[dict] = None,
        tdnn: Optional[dict] = None,
        embedding: Optional[dict] = None,
    ):
        """waveform -> SincNet -> XVectorNet (TDNN -> FC) -> output

        Parameters
        ----------
        sincnet : `dict`, optional
            SincNet parameters. Defaults to `pyannote.audio.models.sincnet.SincNet`
            default parameters.
        tdnn : `dict`, optional
            X-Vector Time-Delay neural network parameters.
            Defaults to `pyannote.audio.models.tdnn.XVectorNet` default parameters.
        embedding : `dict`, optional
            Embedding parameters. Defaults to `Embedding` default parameters. This
            only has effect when model is used for representation learning.
        """

        n_features = self.n_features

        if sincnet is None:
            sincnet = dict()
        self.sincnet = sincnet

        if n_features != 1:
            raise ValueError(
                "SincNet only supports mono waveforms. "
                f"Here, waveform has {n_features} channels."
            )
        self.sincnet_ = SincNet(**sincnet)
        n_features = self.sincnet_.dimension

        if tdnn is None:
            tdnn = dict()
        self.tdnn = tdnn
        self.tdnn_ = XVectorNet(n_features, **tdnn)
        n_features = self.tdnn_.dimension

        if self.task.is_representation_learning:
            if embedding is None:
                embedding = dict()
            self.embedding = embedding
            self.embedding_ = Embedding(n_features, **embedding)
        else:
            self.linear_ = nn.Linear(n_features, len(self.classes), bias=True)
            self.activation_ = self.task.default_activation

    def forward(self, waveforms: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        waveforms : (batch_size, n_samples, 1) `torch.Tensor`
            Batch of waveforms

        Returns
        -------
        output : `torch.Tensor`
            Final network output or intermediate network output
            (only when `return_intermediate` is provided).
        """

        output = self.sincnet_(waveforms)

        return_intermediate = (
            "segment6" if self.task.is_representation_learning else None
        )
        output = self.tdnn_(output, return_intermediate=return_intermediate)

        if self.task.is_representation_learning:
            return self.embedding_(output)

        return self.activation_(self.linear_(output))

    @property
    def dimension(self):
        if self.task.is_representation_learning:
            return self.embedding_.dimension

        return Model.dimension.fget(self)


class ACRoPoLiS(Model):
    """Audio -> Convolutional -> Recurrent (-> optional Pooling) -> Linear -> Scores

    Parameters
    ----------
    specifications : dict
        Task specifications.
    convolutional : dict, optional
        Definition of convolutional layers.
        Defaults to convolutional.Convolutional default hyper-parameters.
    recurrent : dict, optional
        Definition of recurrent layers.
        Defaults to recurrent.Recurrent default hyper-parameters.
    pooling : {"last", ""}, optional
        Definition of pooling layer. Only used when self.task.returns_vector is
        True, in which case it defaults to "last" pooling.
    linear : dict, optional
        Definition of linear layers.
        Defaults to linear.Linear default hyper-parameters.
    scale : dict, optional
    """

    def init(
        self,
        convolutional: dict = None,
        recurrent: dict = None,
        linear: dict = None,
        pooling: Text = None,
        scale: dict = None,
    ):

        self.normalize = nn.InstanceNorm1d(self.n_features)

        if convolutional is None:
            convolutional = dict()
        self.convolutional = convolutional
        self.cnn = Convolutional(self.n_features, **convolutional)

        if recurrent is None:
            recurrent = dict()
        self.recurrent = recurrent
        self.rnn = Recurrent(self.cnn.dimension, **recurrent)

        if pooling is None and self.task.returns_vector:
            pooling = "last"
        if pooling is not None and self.task.returns_sequence:
            msg = f"'pooling' should not be used for labeling tasks (is: {pooling})."
            raise ValueError(msg)

        self.pooling = pooling
        self.pool = Pooling(
            self.rnn.dimension,
            method=self.pooling,
            bidirectional=self.rnn.bidirectional,
        )

        if linear is None:
            linear = dict()
        self.linear = linear
        self.ff = Linear(self.pool.dimension, **linear)

        if self.task.is_representation_learning and scale is None:
            scale = dict()
        if not self.task.is_representation_learning and scale is not None:
            msg = (
                f"'scale' should not be used for representation learning (is: {scale})."
            )
            raise ValueError(msg)

        self.scale = scale
        self.scaling = Scaling(self.ff.dimension, **scale)

        if not self.task.is_representation_learning:
            self.final_linear = nn.Linear(
                self.scaling.dimension, len(self.classes), bias=True
            )
            self.final_activation = self.task.default_activation

    def forward(self, waveforms: torch.Tensor, **kwargs) -> torch.Tensor:
        """Forward pass

        Parameters
        ----------
        waveforms : (batch_size, n_samples, 1) `torch.Tensor`
            Batch of waveforms

        Returns
        -------
        output : `torch.Tensor`
            Final network output or intermediate network output
            (only when `return_intermediate` is provided).
        """

        output = self.normalize(waveforms.transpose(1, 2)).transpose(1, 2)
        output = self.cnn(output)
        output = self.rnn(output)
        output = self.pool(output)
        output = self.ff(output)

        if not self.task.is_representation_learning:
            output = self.final_linear(output)
            output = self.final_activation(output)
        return output

    @property
    def dimension(self):
        if self.task.is_representation_learning:
            return self.ff.dimension
        else:
            return len(self.classes)

    @staticmethod
    def get_alignment(task: Task, convolutional: Optional[dict] = None, **kwargs):
        """Get frame alignment"""

        if convolutional is None:
            convolutional = dict()

        return Convolutional.get_alignment(task, **convolutional)

    @staticmethod
    def get_resolution(
        task: Task, convolutional: Optional[dict] = None, **kwargs
    ) -> Resolution:
        """Get frame resolution

        Parameters
        ----------
        task : Task
        convolutional : dict, optional

        Returns
        -------
        sliding_window : `pyannote.core.SlidingWindow` or {`window`, `frame`}
        """

        if task.returns_vector:
            return RESOLUTION_CHUNK

        if convolutional is None:
            convolutional = dict()

        return Convolutional.get_resolution(task, **convolutional)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Speech activity detection pipelines"""

from typing import Union
import numpy as np
import warnings

from pyannote.pipeline import Pipeline
from pyannote.pipeline.parameter import Uniform

from pyannote.core import Annotation
from pyannote.core import SlidingWindowFeature

from pyannote.audio.utils.signal import Binarize
from pyannote.audio.features import Precomputed

from pyannote.metrics.detection import DetectionErrorRate
from pyannote.metrics.detection import DetectionPrecisionRecallFMeasure
from pyannote.audio.features.wrapper import Wrapper, Wrappable


class OracleSpeechActivityDetection(Pipeline):
    """Oracle speech activity detection"""

    def __call__(self, current_file: dict) -> Annotation:
        """Return groundtruth speech activity detection

        Parameter
        ---------
        current_file : `dict`
            Dictionary as provided by `pyannote.database`.

        Returns
        -------
        hypothesis : `pyannote.core.Annotation`
            Speech regions
        """

        speech = current_file["annotation"].get_timeline().support()
        return speech.to_annotation(generator="string", modality="speech")


class SpeechActivityDetection(Pipeline):
    """Speech activity detection pipeline

    Parameters
    ----------
    scores : Wrappable, optional
        Describes how raw speech activity detection scores should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
        Defaults to "@sad_scores" that indicates that protocol files provide
        the scores in the "sad_scores" key.
    fscore : bool, optional
        Optimize (precision/recall) fscore. Defaults to optimizing detection
        error rate.

    Hyper-parameters
    ----------------
    onset, offset : `float`
        Onset/offset detection thresholds
    min_duration_on, min_duration_off : `float`
        Minimum duration in either state (speech or not)
    pad_onset, pad_offset : `float`
        Padding duration.
    """

    def __init__(self, scores: Wrappable = None, fscore: bool = False):
        super().__init__()

        if scores is None:
            scores = "@sad_scores"
        self.scores = scores
        self._scores = Wrapper(self.scores)

        self.fscore = fscore

        # hyper-parameters
        self.onset = Uniform(0.0, 1.0)
        self.offset = Uniform(0.0, 1.0)
        self.min_duration_on = Uniform(0.0, 2.0)
        self.min_duration_off = Uniform(0.0, 2.0)
        self.pad_onset = Uniform(-1.0, 1.0)
        self.pad_offset = Uniform(-1.0, 1.0)

    def initialize(self):
        """Initialize pipeline with current set of parameters"""

        self._binarize = Binarize(
            onset=self.onset,
            offset=self.offset,
            min_duration_on=self.min_duration_on,
            min_duration_off=self.min_duration_off,
            pad_onset=self.pad_onset,
            pad_offset=self.pad_offset,
        )

    def __call__(self, current_file: dict) -> Annotation:
        """Apply speech activity detection

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol. May contain a
            'sad_scores' key providing precomputed scores.

        Returns
        -------
        speech : `pyannote.core.Annotation`
            Speech regions.
        """

        sad_scores = self._scores(current_file)

        # if this check has not been done yet, do it once and for all
        if not hasattr(self, "log_scale_"):
            # heuristic to determine whether scores are log-scaled
            if np.nanmean(sad_scores.data) < 0:
                self.log_scale_ = True
            else:
                self.log_scale_ = False

        data = np.exp(sad_scores.data) if self.log_scale_ else sad_scores.data

        # speech vs. non-speech
        if data.shape[1] > 1:
            speech_prob = SlidingWindowFeature(
                1.0 - data[:, 0], sad_scores.sliding_window
            )
        else:
            speech_prob = SlidingWindowFeature(data, sad_scores.sliding_window)

        speech = self._binarize.apply(speech_prob)

        speech.uri = current_file.get("uri", None)
        return speech.to_annotation(generator="string", modality="speech")

    def get_metric(
        self, parallel=False
    ) -> Union[DetectionErrorRate, DetectionPrecisionRecallFMeasure]:
        """Return new instance of detection metric"""

        if self.fscore:
            return DetectionPrecisionRecallFMeasure(
                collar=0.0, skip_overlap=False, parallel=parallel
            )
        else:
            return DetectionErrorRate(collar=0.0, skip_overlap=False, parallel=parallel)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional
from typing import Union
from typing import Text
from pathlib import Path

from pyannote.core import Annotation
from pyannote.pipeline import Pipeline
from .speaker_change_detection import SpeakerChangeDetection
from .speech_activity_detection import SpeechActivityDetection
from .speech_activity_detection import OracleSpeechActivityDetection

from pyannote.database import get_annotated
from pyannote.metrics.diarization import DiarizationPurityCoverageFMeasure


class OracleSpeechTurnSegmentation(Pipeline):
    """Oracle segmentation"""

    def __call__(self, current_file: dict) -> Annotation:
        """Return groundtruth segmentation

        Parameter
        ---------
        current_file : `dict`
            Dictionary as provided by `pyannote.database`.

        Returns
        -------
        hypothesis : `pyannote.core.Annotation`
            Speech turns
        """

        return current_file["annotation"].relabel_tracks(generator="string")


class SpeechTurnSegmentation(Pipeline):
    """Combine speech activity and speaker change detections for segmentation

    Parameters
    ----------
    sad_scores : Text or Path or 'oracle', optional
    scd_scores : Text or Path, optional
        Describes how raw speech activity and speaker change detection scores
        should be obtained. It can be either the name of a torch.hub model, or
        the path to the output of the validation step of a model trained
        locally, or the path to scores precomputed on disk. Defaults to
        "@sad_scores" and "@scd_scores" respectively, indicating that protocol
        files provide the scores in the corresponding "sad_scores" and
        "scd_scores" keys. Use 'oracle' to assume perfect speech activity detection.
        Path to precomputed speaker change detection scores
    non_speech : `bool`
        Mark non-speech regions as speaker change. Defaults to True.
    purity : `float`, optional
        Target purity. Defaults to 0.95
    """

    def __init__(
        self,
        sad_scores: Union[Text, Path] = None,
        scd_scores: Union[Text, Path] = None,
        non_speech: Optional[bool] = True,
        purity: Optional[float] = 0.95,
    ):
        super().__init__()

        self.sad_scores = sad_scores
        if self.sad_scores == "oracle":
            self.speech_activity_detection = OracleSpeechActivityDetection()
        else:
            self.speech_activity_detection = SpeechActivityDetection(
                scores=self.sad_scores
            )

        self.scd_scores = scd_scores
        self.speaker_change_detection = SpeakerChangeDetection(scores=self.scd_scores)

        self.non_speech = non_speech
        self.purity = purity

    def __call__(self, current_file: dict) -> Annotation:
        """Apply speech turn segmentation

        Parameter
        ---------
        current_file : `dict`
            Dictionary as provided by `pyannote.database`.

        Returns
        -------
        hypothesis : `pyannote.core.Annotation`
            Hypothesized speech turns
        """

        # speech regions
        sad = self.speech_activity_detection(current_file).get_timeline()

        scd = self.speaker_change_detection(current_file)
        speech_turns = scd.crop(sad, mode="intersection")

        # at this point, consecutive speech turns separated by non-speech
        # might be assigned the same label (because scd might have missed
        # speech/non-speech boundaries)

        # assign one unique label per speech turn
        if self.non_speech:
            speech_turns = speech_turns.relabel_tracks(generator="string")

        speech_turns.modality = "speaker"
        return speech_turns

    def loss(self, current_file: dict, hypothesis: Annotation) -> float:
        """Compute (1 - coverage) at target purity

        If purity < target, return 1 + (1 - purity)

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.
        hypothesis : `pyannote.core.Annotation`
            Speech turns.

        Returns
        -------
        error : `float`
            1. - cluster coverage.
        """

        metric = DiarizationPurityCoverageFMeasure()
        reference = current_file["annotation"]
        uem = get_annotated(current_file)
        f_measure = metric(reference, hypothesis, uem=uem)
        purity, coverage, _ = metric.compute_metrics()
        if purity > self.purity:
            return 1.0 - coverage
        else:
            return 1.0 + (1.0 - purity)
